"""
Metric wrappers are classes which wrap instances of :class:`.Metric` or, in the case of :class:`EpochLambda` and
:class:`BatchLambda`, functions. Typically, these should **not** be used directly (although this is entirely possible),
but via the :mod:`decorator API<.metrics.decorators>`.
"""
import torchbearer
from torchbearer.bases import Metric
from .metrics import AdvancedMetric

import torch


class ToDict(AdvancedMetric):
    """The :class:`ToDict` class is an :class:`.AdvancedMetric` which will put output from the inner :class:`.Metric` in
    a dict (mapping metric name to value) before returning. When in `eval` mode, 'val\_' will be prepended to the metric
    name.

    Example: ::

        >>> from torchbearer import metrics

        >>> @metrics.lambda_metric('my_metric')
        ... def my_metric(y_pred, y_true):
        ...     return y_pred + y_true
        ...
        >>> metric = metrics.ToDict(my_metric().build())
        >>> metric.process({'y_pred': 4, 'y_true': 5})
        {'my_metric': 9}
        >>> metric.eval()
        >>> metric.process({'y_pred': 4, 'y_true': 5})
        {'val_my_metric': 9}

    Args:
        metric (Metric): The :class:`.Metric` instance to *wrap*.
    """

    def __init__(self, metric):
        super(ToDict, self).__init__(metric.name)

        self.eval_flag = 'val'
        self.metric = metric

    def process_train(self, *args):
        val = self.metric.process(*args)
        if val is not None:
            return {self.metric.name: val}

    def process_validate(self, *args):
        val = self.metric.process(*args)
        if val is not None:
            return {self.eval_flag + '_' + self.metric.name: val}

    def process_final_train(self, *args):
        val = self.metric.process_final(*args)
        if val is not None:
            return {self.metric.name: val}

    def process_final_validate(self, *args):
        val = self.metric.process_final(*args)
        if val is not None:
            return {self.eval_flag + '_' + self.metric.name: val}

    def eval(self, data_key=None):
        super(ToDict, self).eval(data_key=data_key)
        if data_key == torchbearer.TEST_DATA:
            self.eval_flag = 'test'
        elif data_key == torchbearer.TRAIN_DATA:
            self.eval_flag = 'train'
        else:
            self.eval_flag = 'val'
        self.metric.eval(data_key=data_key)

    def train(self):
        super(ToDict, self).train()
        self.metric.train()

    def reset(self, state):
        super(ToDict, self).reset(state)
        self.metric.reset(state)


class BatchLambda(Metric):
    """A metric which returns the output of the given function on each batch.

    Args:
        name (str): The name of the metric.
        metric_function (func): A metric function('y_pred', 'y_true') to wrap.
    """

    def __init__(self, name, metric_function):
        super(BatchLambda, self).__init__(name)
        self._metric_function = metric_function

    def process(self, *args):
        """Return the output of the wrapped function.

        Args:
            args: The :class:`.torchbearer.Trial` state.

        Returns:
            The value of the metric function('y_pred', 'y_true').
        """
        state = args[0]
        return self._metric_function(state[torchbearer.Y_PRED], state[torchbearer.Y_TRUE])


class EpochLambda(AdvancedMetric):
    """A metric wrapper which computes the given function for concatenated values of 'y_true' and 'y_pred' each epoch.
    Can be used as a running metric which computes the function for batches of outputs with a given step size during
    training.

    Args:
        name (str): The name of the metric.
        metric_function (func): The function('y_pred', 'y_true') to use as the metric.
        running (bool): True if this should act as a running metric.
        step_size (int): Step size to use between calls if running=True.
    """

    def __init__(self, name, metric_function, running=True, step_size=50):
        super(EpochLambda, self).__init__(name)
        self._step = metric_function
        self._final = metric_function
        self._step_size = step_size
        self._result = 0.0

        if not running:
            self._step = lambda y_pred, y_true: None

    def process_train(self, *args):
        """Concatenate the 'y_true' and 'y_pred' from the state along the 0 dimension, this must be the batch dimension.
        If this is a running metric, evaluates the function every number of steps.

        Args:
            args: The :class:`.torchbearer.Trial` state.

        Returns:
            The current running result.
        """
        state = args[0]
        if (self._y_pred is None) or (self._y_true is None):
            self._y_true = state[torchbearer.Y_TRUE]
            self._y_pred = state[torchbearer.Y_PRED]
        else:
            self._y_true = torch.cat((self._y_true, state[torchbearer.Y_TRUE]), dim=0)
            self._y_pred = torch.cat((self._y_pred, state[torchbearer.Y_PRED]), dim=0)
        if state[torchbearer.BATCH] % self._step_size == 0:
            self._result = self._step(self._y_pred, self._y_true)
        return self._result

    def process_final_train(self, *args):
        """Evaluate the function with the aggregated outputs.

        Returns:
            The result of the function.
        """
        return self._final(self._y_pred, self._y_true)

    def process_validate(self, *args):
        """During validation, just concatenate 'y_true' and y_pred'.

        Args:
            args: The :class:`.torchbearer.Trial` state.
        """
        state = args[0]
        if (self._y_pred is None) or (self._y_true is None):
            self._y_true = state[torchbearer.Y_TRUE]
            self._y_pred = state[torchbearer.Y_PRED]
        else:
            self._y_true = torch.cat((self._y_true, state[torchbearer.Y_TRUE]), dim=0)
            self._y_pred = torch.cat((self._y_pred, state[torchbearer.Y_PRED]), dim=0)

    def process_final_validate(self, *args):
        """Evaluate the function with the aggregated outputs.

        Returns:
            The result of the function.
        """
        return self._final(self._y_pred, self._y_true)

    def reset(self, state):
        """Reset the 'y_true' and 'y_pred' caches.

        Args:
            state (dict): The :class:`.torchbearer.Trial` state.
        """
        super(EpochLambda, self).reset(state)
        self._y_true = None
        self._y_pred = None



from __future__ import print_function
import torchbearer

from torchbearer.callbacks import Callback
from .decorators import only_if


class EarlyStopping(Callback):
    """Callback to stop training when a monitored quantity has stopped improving.

    Example: ::

        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import EarlyStopping

        # Example Trial which does early stopping if the validation accuracy drops below the max seen for 5 epochs in a row
        >>> stopping = EarlyStopping(monitor='val_acc', patience=5, mode='max')
        >>> trial = Trial(None, callbacks=[stopping], metrics=['acc'])

    Args:
        monitor (str): Name of quantity in metrics to be monitored
        min_delta (float): Minimum change in the monitored quantity to qualify as an improvement, i.e. an absolute
            change of less than min_delta, will count as no improvement.
        patience (int): Number of epochs with no improvement after which training will be stopped.
        mode (str): One of {auto, min, max}. In `min` mode, training will stop when the quantity monitored has stopped
            decreasing; in `max` mode it will stop when the quantity monitored has stopped increasing; in `auto` mode,
            the direction is automatically inferred from the name of the monitored quantity.

    State Requirements:
        - :attr:`torchbearer.state.METRICS`: Metrics should be a dict containing the given monitor key as a minimum
    """

    def __init__(self, monitor='val_loss',
                 min_delta=0, patience=0, mode='auto', step_on_batch=False):

        super(EarlyStopping, self).__init__()

        self.monitor = monitor
        self.min_delta = min_delta
        self.patience = patience
        self.mode = mode
        self.step_on_batch = step_on_batch

        if self.mode not in ['min', 'max']:
            if 'acc' in self.monitor:
                self.mode = 'max'
            else:
                self.mode = 'min'

        if self.mode == 'min':
            self.min_delta *= -1
            self.monitor_op = lambda x1, x2: x1 < x2
        elif self.mode == 'max':
            self.min_delta *= 1
            self.monitor_op = lambda x1, x2: x1 > x2

        self.wait = 0
        self.best = float('inf') if self.mode == 'min' else -float('inf')

    def state_dict(self):
        state_dict = {
            'wait': self.wait,
            'best': self.best
        }
        return state_dict

    def load_state_dict(self, state_dict):
        self.wait = state_dict['wait']
        self.best = state_dict['best']

    def step(self, state):
        current = state[torchbearer.METRICS][self.monitor]
        if self.monitor_op(current - self.min_delta, self.best):
            self.best = current
            self.wait = 0
        else:
            self.wait += 1
            if self.wait >= self.patience:
                state[torchbearer.STOP_TRAINING] = True

    @only_if(lambda self, _: self.step_on_batch)
    def on_step_training(self, state):
        self.step(state)

    @only_if(lambda self, _: not self.step_on_batch)
    def on_end_epoch(self, state):
        self.step(state)



from torchbearer import Callback
import torchbearer
from torch.distributions.beta import Beta
import random
import torch
import types
from torchbearer import cite


manifold_mixup = """
@inproceedings{verma2019manifold,
  title={Manifold Mixup: Better Representations by Interpolating Hidden States},
  author={Verma, Vikas and Lamb, Alex and Beckham, Christopher and Najafi, Amir and Mitliagkas, Ioannis and Lopez-Paz, David and Bengio, Yoshua},
  booktitle={International Conference on Machine Learning},
  pages={6438--6447},
  year={2019}
}
"""


@cite(manifold_mixup)
class ManifoldMixup(Callback):
    """ Performs manifold mixup on desired layers. Requires use of :meth:`MixupInputs.loss`, otherwise lambdas can be found in
    state under :attr:`.MIXUP_LAMBDA`. Model targets will be a tuple containing the original target and permuted target.

    Args:
        lam: Mixup lambda. If RANDOM, choose lambda from Beta(alpha, alpha). Else, lambda=lam
        alpha: Mixup alpha. Alpha used to sample lambda from Beta(alpha, alpha)
    """
    RANDOM = -10.0

    def __init__(self, alpha=1.0, lam=RANDOM):
        super(ManifoldMixup, self).__init__()
        self._layers = []
        self._mixup_layers = None
        self.alpha = alpha
        self.lam = lam
        self.distrib = Beta(self.alpha, self.alpha)
        self.layer_names = []
        self.depth = 0
        self._layer_filter = []
        self._layer_types = []

    def _sample_lam(self):
        if self.lam is self.RANDOM:
            if self.alpha > 0:
                lam = self.distrib.sample()
            else:
                lam = 1.0
        else:
            lam = self.lam
        return lam

    def _single_to_list(self, item):
        if not isinstance(item, list) or isinstance(item, tuple):
            item = [item, ]
        return item

    def for_layers(self, layers):
        """ Specify the layer names on which to apply manifold mixup.

        Args:
            layers (list or str): Layer names, eg ['conv1', 'fc1']

        Returns: self
        """
        layers = self._single_to_list(layers)
        self._mixup_layers = layers
        return self

    def with_layer_filter(self, layer_filter=()):
        """ Specify layer names to exclude from manifold mixup.

        Args:
            layer_filter (list or str): Layer filter, eg ['conv1', 'fc1']

        Returns: self
        """
        layer_filter = self._single_to_list(layer_filter)
        self._layer_filter = layer_filter
        return self

    def with_layer_type_filter(self, layer_types=()):
        """ Specify the layer types to exclude from manifold mixup

        Args:
            layer_types (list or str): Layer types, eg [nn.RelU]:

        Returns: self
        """
        layer_types = self._single_to_list(layer_types)
        self._layer_types = layer_types
        return self

    def at_depth(self, N):
        """ Specify the module depth at which to search for layers. Top level modules are at level 0. 
        Submodules of these are at level 1, etc. 
        To search all depths, set N=None.

        Args:
            N (int or None): Module depth

        Returns: self
        """
        self.depth = N
        return self

    def _wrap_layers(self, model, state):
        # Wrap the chosen layers with redefined forward that does mixup
        self._recursive_wrap(model, '', state, 0)

    def on_start(self, state):
        super(ManifoldMixup, self).on_start(state)
        self._wrap_layers(state[torchbearer.MODEL], state)

    def on_sample(self, state):
        lam = self._sample_lam()
        state[torchbearer.MIXUP_LAMBDA] = lam

        layer = random.choice(self._layers)
        layer.mixup()

        state[torchbearer.MIXUP_PERMUTATION] = torch.randperm(state[torchbearer.X].size(0))
        state[torchbearer.Y_TRUE] = (state[torchbearer.Y_TRUE], state[torchbearer.Y_TRUE][state[torchbearer.MIXUP_PERMUTATION]])
        
    def _wrap_layer_check(self, module, name, nname):
        # Check for exclusions
        name_check = self._mixup_layers is None or nname in self._mixup_layers

        filters = [
            any([f == nname for f in self._layer_filter]),
            any([isinstance(module, t) for t in self._layer_types]),
        ]
        return name_check and not any(filters)
    
    def get_selected_layers(self, model):
        layer_names = []
        return self._recursive_name_seach(layer_names, model, '', 0)

    def _recursive_name_seach(self, layer_names, layer, pre_name, depth):
        for name, module in layer.named_children():
            nname = pre_name + '_' + name if pre_name != '' else name
            if depth == self.depth or self.depth is None:
                if self._wrap_layer_check(module, name, nname):
                    layer_names.append(nname)
                
            if self.depth is None or depth <= self.depth:
                if len(list(layer.named_children())) > 0:
                    self._recursive_name_seach(layer_names, module, nname, depth+1)
        return layer_names

    def _recursive_wrap(self, layer, pre_name, state, depth):
        for name, module in layer.named_children():
            nname = pre_name + '_' + name if pre_name != '' else name

            def new_forward(old_forward):
                def new_forward_1(self, *args, **kwargs):
                    o = old_forward(*args, **kwargs)
                    
                    if self.do_mixup and self.training:
                        o = _mixup_inputs(o, state)

                    self.do_mixup = False
                    return o

                return new_forward_1

            if depth == self.depth or self.depth is None:
                self.layer_names.append(nname)

                if self._wrap_layer_check(module, name, nname):
                    self._layers.append(module)
                    nf = new_forward(module.forward)

                    bound_forward = types.MethodType(nf, module)
                    bound_mixup = types.MethodType(_mixup, module)

                    module.__setattr__('forward', bound_forward)
                    module.__setattr__('mixup', bound_mixup)
                    module.__setattr__('state', state)
                    module.__setattr__('do_mixup', False)

            if self.depth is None or depth <= self.depth:
                if len(list(layer.named_children())) > 0:
                    self._recursive_wrap(module, nname, state, depth+1)


def _mixup(self):
    self.do_mixup = True


def _mixup_inputs(x, state):
    index = state[torchbearer.MIXUP_PERMUTATION]
    lam = state[torchbearer.MIXUP_LAMBDA]
    mixed_x = lam * x + (1 - lam) * x[index,:]
    return mixed_x




import torch

import torchbearer
from torchbearer import cite
from torchbearer.callbacks import Callback, only_if, once_per_epoch


bibtex = """
@article{inoue2018data,
  title={Data augmentation by pairing samples for images classification},
  author={Inoue, Hiroshi},
  journal={arXiv preprint arXiv:1801.02929},
  year={2018}
}
"""


@cite(bibtex)
class SamplePairing(Callback):
    """Perform SamplePairing on the model inputs. This is the process of averaging each image with another random image
    without changing the targets. The key here is to use the policy function to only do this some of the time.

    Example: ::

        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import SamplePairing

        # Example Trial which does Sample Pairing regularisation with the policy from the paper
        >>> pairing = SamplePairing()
        >>> trial = Trial(None, criterion=Mixup.loss, callbacks=[pairing], metrics=['acc'])

    Args:
        policy: A function of state which returns True if the current batch should be paired.
    """

    @staticmethod
    def default_policy(start_epoch, end_epoch, on_epochs, off_epochs):
        """Return a policy which performs sample pairing according to the process defined in the paper.

        Args:
            start_epoch (int): Epoch to start pairing on
            end_epoch (int): Epoch to end pairing on (and begin fine-tuning)
            on_epochs (int): Number of epochs to run sample pairing for before a break
            off_epochs (int): Number of epochs to break for

        Returns:
            A policy function
        """
        cache = {'tick': 0, 'on': False}

        @once_per_epoch
        def compute(state):
            if start_epoch <= state[torchbearer.EPOCH] < end_epoch:
                if cache['tick'] < on_epochs:
                    cache['on'] = True
                else:
                    cache['on'] = False

                cache['tick'] = cache['tick'] + 1
                if cache['tick'] == (on_epochs + off_epochs):
                    cache['tick'] = 0
                    cache['on'] = False
            else:
                cache['on'] = False

        def policy(state):
            compute(state)
            return cache['on']

        return policy

    def __init__(self, policy=None):
        self.on_sample = only_if(SamplePairing.default_policy(100, 800, 8, 2) if policy is None else policy)(self.on_sample)

    def on_sample(self, state):
        permutation = torch.randperm(state[torchbearer.X].size(0))
        state[torchbearer.INPUT] = (state[torchbearer.INPUT] + state[torchbearer.INPUT][permutation]) / 2.



import sys

if sys.version_info[0] < 3:
    import inspect
    def get_default(fcn, arg):
        a = inspect.getargspec(fcn)
        return dict(zip(a.args[-len(a.defaults):], a.defaults))[arg]
else:
    from inspect import signature
    def get_default(fcn, arg):
        return signature(fcn).parameters[arg].default

import functools
import warnings
import itertools

import torch
import torch.nn
from torch.utils.data import DataLoader, TensorDataset
from torch.optim import Optimizer

import torchbearer
from torchbearer import cite
from torchbearer import State
from torchbearer.metrics import MetricList
from torchbearer.callbacks import Callback, CallbackList, Tqdm, AggregatePredictions
from torchbearer.bases import standard_closure, _forward_with_exceptions, _get_param_list

bibtex = """
@article{2018torchbearer,
  title={Torchbearer: A Model Fitting Library for PyTorch},
  author={Harris, Ethan and Painter, Matthew and Hare, Jonathon},
  journal={arXiv preprint arXiv:1809.03363},
  year={2018}
}
"""


class MockOptimizer(Optimizer):
    """The Mock Optimizer will be used inplace of an optimizer in the event that none is passed to the Trial class.
    """

    def __init__(self):
        super(MockOptimizer, self).__init__([torch.zeros(1)], [])

    def add_param_group(self, param_group):
        pass  # Do Nothing

    def load_state_dict(self, state_dict):
        pass  # Do Nothing

    def state_dict(self):
        return {}  # Return Empty

    def step(self, closure=None):
        if closure is not None:
            closure()

    def zero_grad(self):
        pass  # Do Nothing


class MockModel(torch.nn.Module):
    def forward(self, x, state=None):
        return None


class CallbackListInjection(CallbackList):
    """This class allows for an callback to be injected into a callback list, without masking the methods available for
    mutating the list. In this way, callbacks (such as printers) can be injected seamlessly into the methods of the
    trial class.

    Args:
        callback (Callback): The :class:`.Callback` to inject
        callback_list (CallbackList): The underlying :class:`.CallbackList`
    """

    def __init__(self, callback, callback_list):
        super(CallbackListInjection, self).__init__([])

        self.callback = callback
        self.callback_list = callback_list

    def state_dict(self):
        return self.callback_list.state_dict()

    def load_state_dict(self, state_dict):
        self.callback_list.load_state_dict(state_dict)
        return self

    def __iter__(self):
        return self.callback_list.__iter__()

    def __copy__(self):
        return self.callback_list.copy()

    def copy(self):
        return self.__copy__()

    def append(self, callback_list):
        self.callback_list.append(callback_list)

    def _for_list(self, function):
        function(self.callback)  # Call injected callback BEFORE the callback list
        function(self.callback_list)


def inject_printer(validation_label_letter='v'):
    """The inject printer decorator is used to inject the appropriate printer callback, according to the verbosity level.

    Args:
        validation_label_letter (str): The validation label letter to use

    Returns:
        A decorator
    """
    from inspect import getcallargs

    def decorator(func):
        root = func if not hasattr(func, 'root') else func.root
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            call_args = getcallargs(root, self, *args, **kwargs)
            verbose = call_args['verbose'] if 'verbose' in call_args else get_default(func, 'verbose')  # Populate default value
            verbose = self.verbose if verbose == -1 else verbose

            printer = get_printer(verbose=verbose, validation_label_letter=validation_label_letter)

            callback_list_old = self.state[torchbearer.CALLBACK_LIST]

            self.state[torchbearer.CALLBACK_LIST] = CallbackListInjection(printer, callback_list_old)

            res = func(self, *args, **kwargs)

            self.state[torchbearer.CALLBACK_LIST] = callback_list_old
            return res
        wrapper.root = root
        return wrapper
    return decorator


def get_printer(verbose, validation_label_letter):
    if verbose >= 2:
        printer = Tqdm(validation_label_letter=validation_label_letter)
    elif verbose >= 1:
        printer = Tqdm(validation_label_letter=validation_label_letter, on_epoch=True)
    else:
        printer = Callback()

    return printer


def deep_to(batch, device, dtype):
    """ Static method to call :func:`to` on tensors, tuples or dicts. All items will have :func:`deep_to` called

    Example: ::

        >>> import torch
        >>> from torchbearer import deep_to
        >>> example_dict = {'a': torch.ones(5)*2.1, 'b': torch.ones(1)*5.9}
        >>> deep_to(example_dict, device='cpu', dtype=torch.int)
        {'a': tensor([2, 2, 2, 2, 2], dtype=torch.int32), 'b': tensor([5], dtype=torch.int32)}

    Args:
        batch (tuple / list / torch.Tensor / dict): The mini-batch which requires a :func:`to` call
        device (torch.device): The desired device of the batch
        dtype (torch.dtype): The desired datatype of the batch

    Returns:
        tuple / list / torch.Tensor: The moved or casted batch
    """
    is_tuple = isinstance(batch, tuple)

    if isinstance(batch, list) or isinstance(batch, tuple):
        batch = list(batch)
        for i in range(len(batch)):
            batch[i] = deep_to(batch[i], device, dtype)
        batch = tuple(batch) if is_tuple else batch
    elif isinstance(batch, dict):
        for key in batch:
            batch[key] = deep_to(batch[key], device, dtype)
    elif torch.is_tensor(batch):
        if batch.dtype.is_floating_point:
            batch = batch.to(device, dtype)
        else:
            batch = batch.to(device)

    return batch


def load_batch_infinite(loader):
    """ Wraps a batch loader and refreshes the iterator once it has been completed.

    Args:
        loader: batch loader to wrap
    """

    def call(state):
        try:
            loader(state)
        except StopIteration:
            state[torchbearer.ITERATOR] = iter(state[torchbearer.GENERATOR])
            loader(state)

    return call


def load_batch_standard(state):
    """ Load a standard (input data, target) tuple mini-batch from iterator into state

    Args:
        state (dict): The current state dict of the :class:`Trial`.
    """
    state[torchbearer.X], state[torchbearer.Y_TRUE] = deep_to(next(state[torchbearer.ITERATOR]),
                                                              state[torchbearer.DEVICE],
                                                              state[torchbearer.DATA_TYPE])


def load_batch_none(state):
    """ Load a none (none, none) tuple mini-batch into state

    Args:
        state (dict): The current state dict of the :class:`Trial`.
    """
    state[torchbearer.X], state[torchbearer.Y_TRUE] = None, None


def load_batch_predict(state):
    """ Load a prediction (input data, target) or (input data) mini-batch from iterator into state

    Args:
        state (dict): The current state dict of the :class:`Trial`.
    """
    data = deep_to(next(state[torchbearer.ITERATOR]), state[torchbearer.DEVICE], state[torchbearer.DATA_TYPE])
    if isinstance(data, list) or isinstance(data, tuple):
        try:
            state[torchbearer.X], state[torchbearer.Y_TRUE] = data
        except ValueError:
            state[torchbearer.X] = data[0]
    else:
        state[torchbearer.X] = data


def inject_sampler(data_key, batch_sampler):
    """ Decorator to inject a :class:`Sampler` into state[torchbearer.SAMPLER] along with the specified \
        generator into state[torchbearer.GENERATOR] and number of steps into state[torchbearer.STEPS]

    Args:
        data_key (:class:`.StateKey`): Key for the data to inject
        batch_sampler (function): Batch sampler function that extracts batch from data loader, stores in state and sends
        data to correct device

    Returns:
        The decorator
    """
    from inspect import getcallargs

    def decorator(func):
        root = func if not hasattr(func, 'root') else func.root

        def infinite_wrapper(self, key, generator, steps, sampler):
            if generator is not None and steps is not None:
                over_steps = steps > len(generator)
                inf_steps = steps == -1
                inf_train_loader = key == torchbearer.TRAIN_DATA and self.state[torchbearer.INF_TRAIN_LOADING]

                if over_steps or inf_steps or inf_train_loader:  # Want iterator to refresh at end not per epoch
                    if steps == -1: warnings.warn("Trial is set to run indefinitely. "
                                              "Make sure you have some method to terminate safely.")
                    sampler = load_batch_infinite(sampler)

                # Want iterator to run until end before refreshing regardless of number of train/val steps
                if inf_train_loader and not hasattr(generator, 'tb_iter'):
                    generator.tb_iter = iter(generator)

            return generator, sampler

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            sampler = batch_sampler

            call_args = getcallargs(root, self, *args, **kwargs)
            key = call_args['data_key'] if ('data_key' in call_args and call_args['data_key'] is not None) else data_key  # Populate default value
            generator, steps = self.state[key] if self.state[key] is not None else (None, None)

            if self.state[torchbearer.LOADER] is not None:
                sampler = self.state[torchbearer.LOADER]
            elif generator is None:
                sampler = load_batch_none

            generator, sampler = infinite_wrapper(self, key, generator, steps, sampler)

            self.state[torchbearer.DATA] = key
            self.state[torchbearer.SAMPLER] = sampler
            self.state[torchbearer.GENERATOR] = generator
            self.state[torchbearer.STEPS] = steps

            res = func(self, *args, **kwargs)

            return res
        wrapper.root = root
        return wrapper
    return decorator


def inject_callback(callback):
    """ Decorator to inject a callback into the callback list and remove the callback after the decorated function has executed

    Args:
        callback (Callback): :class:`.Callback` to be injected

    Returns:
        The decorator
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            callback_list_old = self.state[torchbearer.CALLBACK_LIST]

            self.state[torchbearer.CALLBACK_LIST] = CallbackListInjection(callback, callback_list_old)

            res = func(self, *args, **kwargs)

            self.state[torchbearer.CALLBACK_LIST] = callback_list_old
            return res
        return wrapper
    return decorator


def update_device_and_dtype(state, *args, **kwargs):
    """Function gets data type and device values from the args / kwargs and updates state.

    Args:
        state (State): The :class:`.State` to update
        args: Arguments to the :func:`Trial.to` function
        kwargs: Keyword arguments to the :func:`Trial.to` function

    Returns:
        state
    """
    for key, _ in kwargs.items():
        if key == str(torchbearer.DATA_TYPE):
            state[torchbearer.DATA_TYPE] = kwargs['dtype']
        elif str(torchbearer.DEVICE) in kwargs:
            state[torchbearer.DEVICE] = kwargs['device']

    for arg in args:
        if isinstance(arg, torch.dtype):
            state[torchbearer.DATA_TYPE] = arg
        else:
            state[torchbearer.DEVICE] = arg

    return state


@cite(bibtex)
class Trial(object):
    """
    The trial class contains all of the required hyper-parameters for model running in torchbearer and presents an
    API for model fitting, evaluating and predicting.

    Example: ::

        >>> import torch
        >>> from torchbearer import Trial

        # Example trial that attempts to minimise the output of a linear layer.
        # Makes use of a callback to input the random data at each batch and a loss that is the absolute value of the
        # linear layer output. Runs for 10 iterations and a single epoch.
        >>> model = torch.nn.Linear(2,1)
        >>> optimiser = torch.optim.Adam(model.parameters(), lr=3e-4)

        >>> @torchbearer.callbacks.on_sample
        ... def initial_data(state):
        ...     state[torchbearer.X] = torch.rand(1, 2)*10
        >>> def minimise_output_loss(y_pred, y_true):
        ...     return torch.abs(y_pred)
        >>> trial = Trial(model, optimiser, minimise_output_loss, ['loss'], [initial_data]).for_steps(10).run(1)

    Args:
        model (torch.nn.Module): The base pytorch model
        optimizer (torch.optim.Optimizer): The optimizer used for pytorch model weight updates
        criterion (func / None): The final loss criterion that provides a loss value to the optimizer
        metrics (list): The list of :class:`torchbearer.Metric <.Metric>` instances to process during fitting
        callbacks (list): The list of :class:`torchbearer.Callback <.Callback>` instances to call during fitting
        verbose (int): Global verbosity .If 2: use tqdm on batch, If 1: use tqdm on epoch, If 0: display no training
            progress
    """
    def __init__(self, model, optimizer=None, criterion=None, metrics=[], callbacks=[], verbose=2):
        if criterion is None:
            def criterion(_, __):
                return torch.zeros(1, device=self.state[torchbearer.DEVICE], dtype=self.state[torchbearer.DATA_TYPE], requires_grad=True)

        self.verbose = verbose

        self.closure = standard_closure()
        self.state = State()
        self.state.update({
            torchbearer.MODEL: model if model is not None else MockModel(),
            torchbearer.CRITERION: criterion,
            torchbearer.OPTIMIZER: optimizer if optimizer is not None else MockOptimizer(),
            torchbearer.METRIC_LIST: MetricList(metrics),
            torchbearer.CALLBACK_LIST: CallbackList(callbacks),
            torchbearer.DEVICE: 'cpu',
            torchbearer.DATA_TYPE: torch.float32,
            torchbearer.SELF: self,
            torchbearer.HISTORY: [],
            torchbearer.BACKWARD_ARGS: {},
            torchbearer.TRAIN_GENERATOR: None,
            torchbearer.VALIDATION_GENERATOR: None,
            torchbearer.TEST_GENERATOR: None,
            torchbearer.TRAIN_STEPS: None,
            torchbearer.VALIDATION_STEPS: None,
            torchbearer.TEST_STEPS: None,
            torchbearer.TRAIN_DATA: None,
            torchbearer.VALIDATION_DATA: None,
            torchbearer.TEST_DATA: None,
            torchbearer.INF_TRAIN_LOADING: False,
            torchbearer.LOADER: None
        })

        self.state[torchbearer.CALLBACK_LIST].on_init(self.state)

    def __str__(self):
        def state_string(name, state_key):
            import math
            N = (50-len(name))/2
            res = "-" * int(math.floor(N)) + " " + name.upper() + " " + "-" * int(math.ceil(N))
            res = res + "-" if len(res) < 52 else res
            return res + "\n" + str(self.state[state_key]) + "\n\n"

        optim_str = state_string('Optimzer', torchbearer.OPTIMIZER)
        crit_str = state_string("Criterion", torchbearer.CRITERION)
        metrics_str = state_string("Metrics", torchbearer.METRIC_LIST)
        callbacks_str = state_string("Callbacks", torchbearer.CALLBACK_LIST)
        model_str = state_string("Model", torchbearer.MODEL)

        return optim_str + crit_str + metrics_str + callbacks_str + model_str

    def __repr__(self):
        return str(self)

    # Data addition

    def for_train_steps(self, steps):
        """Run this trial for the given number of training steps. Note that the generator will output (None, None) if it
        has not been set. Useful for differentiable programming. Returns self so that methods can be chained for
        convenience. If steps is larger than dataset size then loader will be refreshed like if it was a new epoch. If
        steps is -1 then loader will be refreshed until stopped by STOP_TRAINING flag or similar.

        Example: ::

            # Simple trial that runs for 100 training iterations, in this case optimising nothing
            >>> from torchbearer import Trial
            >>> trial = Trial(None).for_train_steps(100)

        Args:
            steps (int): The number of training steps per epoch to run.

        Returns:
            Trial: self
        """
        if not isinstance(steps, int):
            warnings.warn("Number of training steps is not an int, casting to int")
            steps = int(steps)
        self.state[torchbearer.TRAIN_STEPS] = steps
        self.state[torchbearer.TRAIN_DATA] = (self.state[torchbearer.TRAIN_GENERATOR], self.state[torchbearer.TRAIN_STEPS])

        return self

    def with_train_generator(self, generator, steps=None):
        """Use this trial with the given train generator. Returns self so that methods can be chained for convenience.

        Example: ::

            # Simple trial that runs for 100 training iterations on the MNIST dataset
            >>> from torchbearer import Trial
            >>> from torchvision.datasets import MNIST
            >>> from torch.utils.data import DataLoader
            >>> dataloader = DataLoader(MNIST('./data/', train=True))
            >>> trial = Trial(None).with_train_generator(dataloader).for_steps(100).run(1)

        Args:
            generator: The train data generator to use during calls to :meth:`.run`
            steps (int): The number of steps per epoch to take when using this generator.

        Returns:
            Trial: self
        """
        self.state[torchbearer.TRAIN_GENERATOR] = generator
        steps = self.state[torchbearer.TRAIN_STEPS] if steps is None else steps
        steps = len(generator) if steps is None else steps
        self.for_train_steps(steps)

        return self

    def with_train_data(self, x, y, batch_size=1, shuffle=True, num_workers=1, steps=None):
        """Use this trial with the given train data. Returns self so that methods can be chained for convenience.

        Example: ::

            # Simple trial that runs for 10 training iterations on some random data
            >>> from torchbearer import Trial
            >>> data = torch.rand(10, 1)
            >>> targets = torch.rand(10, 1)
            >>> trial = Trial(None).with_val_data(data, targets).for_steps(10).run(1)

        Args:
            x (torch.Tensor): The train x data to use during calls to :meth:`.run`
            y (torch.Tensor): The train labels to use during calls to :meth:`.run`
            batch_size (int): The size of each batch to sample from the data
            shuffle (bool): If True, then data will be shuffled each epoch
            num_workers (int): Number of worker threads to use in the data loader
            steps (int): The number of steps per epoch to take when using this data

        Returns:
            Trial: self
        """
        dataset = TensorDataset(x, y)
        dataloader = DataLoader(dataset, batch_size, shuffle=shuffle, num_workers=num_workers)
        self.with_train_generator(dataloader, steps=steps)

        return self

    def for_val_steps(self, steps):
        """Run this trial for the given number of validation steps. Note that the generator will output (None, None) if
        it has not been set. Useful for differentiable programming. Returns self so that methods can be chained for
        convenience. If steps larger than dataset size then loader will be refreshed like if it was a new epoch. If
        steps -1 then loader will be refreshed until stopped by STOP_TRAINING flag or similar.

        Example: ::

            # Simple trial that runs for 10 validation iterations on no data
            >>> from torchbearer import Trial
            >>> data = torch.rand(10, 1)
            >>> trial = Trial(None).for_val_steps(10).run(1)

        Args:
            steps (int): The number of validation steps per epoch to run

        Returns:
            Trial: self
        """
        if not isinstance(steps, int):
            warnings.warn("Number of validation steps is not an int, casting to int")
            steps = int(steps)
        self.state[torchbearer.VALIDATION_STEPS] = steps
        self.state[torchbearer.VALIDATION_DATA] = (self.state[torchbearer.VALIDATION_GENERATOR], self.state[torchbearer.VALIDATION_STEPS])

        return self

    def with_val_generator(self, generator, steps=None):
        """Use this trial with the given validation generator. Returns self so that methods can be chained for
        convenience.

        Example: ::

            # Simple trial that runs for 100 validation iterations on the MNIST dataset
            >>> from torchbearer import Trial
            >>> from torchvision.datasets import MNIST
            >>> from torch.utils.data import DataLoader
            >>> dataloader = DataLoader(MNIST('./data/', train=False))
            >>> trial = Trial(None).with_val_generator(dataloader).for_steps(100).run(1)

        Args:
            generator: The validation data generator to use during calls to :meth:`.run` and :meth:`.evaluate`
            steps (int): The number of steps per epoch to take when using this generator

        Returns:
            Trial: self
        """
        self.state[torchbearer.VALIDATION_GENERATOR] = generator
        steps = self.state[torchbearer.VALIDATION_STEPS] if steps is None else steps
        steps = len(generator) if steps is None else steps
        self.for_val_steps(steps)

        return self

    def with_val_data(self, x, y, batch_size=1, shuffle=True, num_workers=1, steps=None):
        """Use this trial with the given validation data. Returns self so that methods can be chained for convenience.

        Example: ::

            # Simple trial that runs for 10 validation iterations on some random data
            >>> from torchbearer import Trial
            >>> data = torch.rand(10, 1)
            >>> targets = torch.rand(10, 1)
            >>> trial = Trial(None).with_val_data(data, targets).for_steps(10).run(1)

        Args:
            x (torch.Tensor): The validation x data to use during calls to :meth:`.run` and :meth:`.evaluate`
            y (torch.Tensor): The validation labels to use during calls to :meth:`.run` and :meth:`.evaluate`
            batch_size (int): The size of each batch to sample from the data
            shuffle (bool): If True, then data will be shuffled each epoch
            num_workers (int): Number of worker threads to use in the data loader
            steps (int): The number of steps per epoch to take when using this data

        Returns:
            Trial: self
        """
        dataset = TensorDataset(x, y)
        dataloader = DataLoader(dataset, batch_size, shuffle=shuffle, num_workers=num_workers)
        self.with_val_generator(dataloader, steps=steps)

        return self

    def for_test_steps(self, steps):
        """Run this trial for the given number of test steps. Note that the generator will output (None, None) if
        it has not been set. Useful for differentiable programming. Returns self so that methods can be chained for
        convenience. If steps larger than dataset size then loader will be refreshed like if it was a new epoch. If
        steps -1 then loader will be refreshed until stopped by STOP_TRAINING flag or similar.

        Example: ::

            # Simple trial that runs for 10 test iterations on some random data
            >>> from torchbearer import Trial
            >>> data = torch.rand(10, 1)
            >>> trial = Trial(None).with_test_data(data).for_test_steps(10).run(1)

        Args:
            steps (int): The number of test steps per epoch to run (when using :meth:`.predict`)

        Returns:
            Trial: self
        """
        if not isinstance(steps, int):
            warnings.warn("Number of test steps is not an int, casting to int")
            steps = int(steps)
        self.state[torchbearer.TEST_STEPS] = steps
        self.state[torchbearer.TEST_DATA] = (self.state[torchbearer.TEST_GENERATOR], self.state[torchbearer.TEST_STEPS])

        return self

    def with_test_generator(self, generator, steps=None):
        """Use this trial with the given test generator. Returns self so that methods can be chained for convenience.

        Example: ::

            # Simple trial that runs for 10 test iterations on no data
            >>> from torchbearer import Trial
            >>> data = torch.rand(10, 1)
            >>> trial = Trial(None).with_test_data(data).for_test_steps(10).run(1)

        Args:
            generator: The test data generator to use during calls to :meth:`.predict`
            steps (int): The number of steps per epoch to take when using this generator

        Returns:
            Trial: self
        """
        self.state[torchbearer.TEST_GENERATOR] = generator
        steps = self.state[torchbearer.TEST_STEPS] if steps is None else steps
        steps = len(generator) if steps is None else steps
        self.for_test_steps(steps)

        return self

    def with_test_data(self, x, batch_size=1, num_workers=1, steps=None):
        """Use this trial with the given test data. Returns self so that methods can be chained for convenience.

        Example: ::

            # Simple trial that runs for 10 test iterations on some random data
            >>> from torchbearer import Trial
            >>> data = torch.rand(10, 1)
            >>> trial = Trial(None).with_test_data(data).for_test_steps(10).run(1)

        Args:
            x (torch.Tensor): The test x data to use during calls to :meth:`.predict`
            batch_size (int): The size of each batch to sample from the data
            num_workers (int): Number of worker threads to use in the data loader
            steps (int): The number of steps per epoch to take when using this data

        Returns:
            Trial: self
        """
        dataset = TensorDataset(x)
        dataloader = DataLoader(dataset, batch_size, num_workers=num_workers)
        self.with_test_generator(dataloader, steps=steps)

        return self

    def for_steps(self, train_steps=None, val_steps=None, test_steps=None):
        """Use this trial for the given number of train, val and test steps. Returns self so that methods can be chained
        for convenience. If steps larger than dataset size then loader will be refreshed like if it was a new epoch. If
        steps -1 then loader will be refreshed until stopped by STOP_TRAINING flag or similar.

        Example: ::

            # Simple trial that runs for 10 training, validation and test iterations on some random data
            >>> from torchbearer import Trial
            >>> train_data = torch.rand(10, 1)
            >>> val_data = torch.rand(10, 1)
            >>> test_data = torch.rand(10, 1)
            >>> trial = Trial(None).with_train_data(train_data).with_val_data(val_data).with_test_data(test_data)
            >>> trial.for_steps(10, 10, 10).run(1)

        Args:
            train_steps (int): The number of training steps per epoch to run
            val_steps (int): The number of validation steps per epoch to run
            test_steps (int): The number of test steps per epoch to run (when using :meth:`.predict`)

        Returns:
            Trial: self
        """
        if train_steps is not None:
            self.for_train_steps(train_steps)
        if val_steps is not None:
            self.for_val_steps(val_steps)
        if test_steps is not None:
            self.for_test_steps(test_steps)

        return self

    def with_generators(self, train_generator=None, val_generator=None, test_generator=None, train_steps=None, val_steps=None, test_steps=None):
        """Use this trial with the given generators. Returns self so that methods can be chained for convenience.

        Example: ::

            # Simple trial that runs for 100 steps from a training and validation data generator
            >>> from torchbearer import Trial
            >>> from torchvision.datasets import MNIST
            >>> from torch.utils.data import DataLoader
            >>> trainloader = DataLoader(MNIST('./data/', train=True))
            >>> valloader = DataLoader(MNIST('./data/', train=False))
            >>> trial = Trial(None).with_generators(trainloader, valloader, train_steps=100, val_steps=100).run(1)

        Args:
            train_generator: The training data generator to use during calls to :meth:`.run`
            val_generator: The validation data generator to use during calls to :meth:`.run` and :meth:`.evaluate`
            test_generator: The testing data generator to use during calls to :meth:`.predict`
            train_steps (int): The number of steps per epoch to take when using the training generator
            val_steps (int): The number of steps per epoch to take when using the validation generator
            test_steps (int): The number of steps per epoch to take when using the testing generator

        Returns:
            Trial: self
        """
        if train_generator is not None:
            self.with_train_generator(train_generator, train_steps)
        if val_generator is not None:
            self.with_val_generator(val_generator, val_steps)
        if test_generator is not None:
            self.with_test_generator(test_generator, test_steps)

        return self

    def with_data(self, x_train=None, y_train=None, x_val=None, y_val=None, x_test=None, batch_size=1,
                  num_workers=1, train_steps=None, val_steps=None, test_steps=None, shuffle=True):
        """Use this trial with the given data. Returns self so that methods can be chained for convenience.

        Example: ::

            # Simple trial that runs for 10 test iterations on some random data
            >>> from torchbearer import Trial
            >>> data = torch.rand(10, 1)
            >>> targets = torch.rand(10, 1)
            >>> test_data = torch.rand(10, 1)
            >>> trial = Trial(None).with_data(x_train=data, y_train=targets, x_test=test_data)
            >>> trial.for_test_steps(10).run(1)

        Args:
            x_train (torch.Tensor): The training data to use
            y_train (torch.Tensor): The training targets to use
            x_val (torch.Tensor): The validation data to use
            y_val (torch.Tensor): The validation targets to use
            x_test (torch.Tensor): The test data to use
            batch_size (int): Batch size to use in mini-batching
            num_workers (int): Number of workers to use for data loading and batching
            train_steps (int): Number of steps for each training pass
            val_steps (int): Number of steps for each validation pass
            test_steps (int): Number of steps for each test pass
            shuffle (bool): If True, shuffle training and validation data.

        Returns:
            Trial: self
        """
        self.with_train_data(x_train, y_train, batch_size, shuffle, num_workers, train_steps)
        self.with_val_data(x_val, y_val, batch_size, shuffle, num_workers, val_steps)
        self.with_test_data(x_test, batch_size, num_workers, test_steps)

    # Infinite steps and loading

    def for_inf_train_steps(self):
        """Use this trial with an infinite number of training steps (until stopped via STOP_TRAINING flag or similar). 
        Returns self so that methods can be chained for convenience.

        Example: ::

            # Simple trial that runs training data until stopped
            >>> from torchbearer import Trial
            >>> from torchvision.datasets import MNIST
            >>> from torch.utils.data import DataLoader
            >>> trainloader = DataLoader(MNIST('./data/', train=True))
            >>> trial = Trial(None).with_train_generator(trainloader).for_inf_train_steps()
            >>> trial.run(1)

        Returns:
            Trial: self
        """
        self.for_train_steps(-1)
        return self

    def for_inf_val_steps(self):
        """Use this trial with an infinite number of validation steps (until stopped via STOP_TRAINING flag or similar).
        Returns self so that methods can be chained for convenience.

        Example: ::

            # Simple trial that runs validation data until stopped
            >>> from torchbearer import Trial
            >>> from torchvision.datasets import MNIST
            >>> from torch.utils.data import DataLoader
            >>> valloader = DataLoader(MNIST('./data/', train=False))
            >>> trial = Trial(None).with_val_generator(valloader).for_inf_val_steps()
            >>> trial.run(1)

        Returns:
            Trial: self
        """
        self.for_val_steps(-1)
        return self

    def for_inf_test_steps(self):
        """Use this trial with an infinite number of test steps (until stopped via STOP_TRAINING flag or similar). 
        Returns self so that methods can be chained for convenience.

        Example: ::

            # Simple trial that runs test data until stopped
            >>> from torchbearer import Trial
            >>> test_data = torch.rand(1000, 10)
            >>> trial = Trial(None).with_test_data(test_data).for_inf_test_steps()
            >>> trial.run(1)

        Returns:
            Trial: self
        """
        self.for_test_steps(-1)
        return self

    def for_inf_steps(self, train=True, val=True, test=True):
        """Use this trail with infinite steps. Returns self so that methods can be chained for convenience.

        Example: ::

            # Simple trial that runs training and test data until stopped
            >>> from torchbearer import Trial
            >>> from torchvision.datasets import MNIST
            >>> from torch.utils.data import DataLoader
            >>> trainloader = DataLoader(MNIST('./data/', train=True))
            >>> valloader = DataLoader(MNIST('./data/', train=False))
            >>> trial = Trial(None).with_train_generator(trainloader).for_inf_steps(valloader)
            >>> trial.with_inf_test_loader(True, False, True).run(1)

        Args:
            train (bool): Use an infinite number of training steps
            val (bool): Use an infinite number of validation steps
            test (bool): Use an infinite number of test steps

        Returns:
            Trial: self
        """
        if train: self.for_inf_train_steps()
        if val: self.for_inf_val_steps()
        if test: self.for_inf_test_steps()

        return self

    def with_inf_train_loader(self):
        """Use this trial with a training iterator that refreshes when it finishes instead of each epoch. 
        This allows for setting training steps less than the size of the generator and model will still be trained on 
        all training samples if enough "epochs" are run.

        Example: ::

            # Simple trial that runs 10 epochs of 100 iterations of a training generator without reshuffling until all data has been seen
            >>> from torchbearer import Trial
            >>> from torchvision.datasets import MNIST
            >>> from torch.utils.data import DataLoader
            >>> trainloader = DataLoader(MNIST('./data/', train=True))
            >>> trial = Trial(None).with_train_generator(trainloader).with_inf_train_loader()
            >>> trial.run(10)

        Returns:
            Trial: self:
        """
        self.state[torchbearer.INF_TRAIN_LOADING] = True

        return self

    # Customise training loop

    def with_loader(self, batch_loader):
        """Use this trial with custom batch loader. Usually calls next on state[torchbearer.ITERATOR] and populates
        state[torchbearer.X] and state[torchbearer.Y_TRUE]

        Example: ::

            # Simple trial that runs with a custom loader function that populates X and Y_TRUE in state with random data
            >>> from torchbearer import Trial
            >>> def custom_loader(state):
            ...     state[X], state[Y_TRUE] = torch.rand(5, 5), torch.rand(5, 5)
            >>> trial = Trial(None).with_loader(custom_loader)
            >>> trial.run(10)

        Args:
            batch_loader (function): Function of state that extracts data from data loader (stored under
                torchbearer.ITERATOR), stores it in state and sends it to the correct device

        Returns:
            Trial: self:
        """
        self.state[torchbearer.LOADER] = batch_loader
        return self

    def with_closure(self, closure):
        """Use this trial with custom closure

        Example: ::

            # Simple trial that runs with a custom closure
            >>> from torchbearer import Trial
            >>> def custom_closure(state):
            ...     print(state[torchbearer.BATCH])
            >>> trial = Trial(None).with_closure(custom_closure).for_steps(3)
            >>> _ = trial.run(1)
            0
            1
            2

        Args:
            closure (function): Function of state that defines the custom closure

        Returns:
            Trial: self:
        """
        self.closure = closure

        return self

    # Run

    @inject_printer()
    def run(self, epochs=1, verbose=-1):
        r"""Run this trial for the given number of epochs, starting from the last trained epoch.

        Example: ::

            # Simple trial that runs with a custom closure
            >>> from torchbearer import Trial
            >>> trial = Trial(None).for_steps(100)
            >>> _ = trial.run(1)

        Args:
            epochs (int, optional): The number of epochs to run for
            verbose (int, optional): If 2: use tqdm on batch, If 1: use tqdm on epoch, If 0: display no training
                progress, If -1: Automatic

        State Requirements:
            - :attr:`torchbearer.state.MODEL`: Model should be callable and not none, set on Trial init

        Returns:
            list: The model history (list of tuple of steps summary and epoch metric dicts)
        """
        state = State()
        state.update({
            torchbearer.MAX_EPOCHS: epochs,
            torchbearer.STOP_TRAINING: False,
        })

        if self.state[torchbearer.MODEL] is None or not callable(self.state[torchbearer.MODEL]):
            warnings.warn('The Model is None or not callable which may cause issues if not deliberate')
            self.state[torchbearer.MODEL] = MockModel()

        state.update(self.state)  # TODO: Swap this for something which makes `self.state` still mutable

        if state[torchbearer.TRAIN_GENERATOR] is not None \
                or state[torchbearer.TRAIN_STEPS] is not None \
                or state[torchbearer.VALIDATION_GENERATOR] is not None \
                or state[torchbearer.VALIDATION_STEPS] is not None:

            state[torchbearer.CALLBACK_LIST].on_start(state)

            for state[torchbearer.EPOCH] in range(len(state[torchbearer.HISTORY]), state[torchbearer.MAX_EPOCHS]):
                state[torchbearer.CALLBACK_LIST].on_start_epoch(state)

                final_metrics = self._fit_pass(state)[torchbearer.METRICS]

                if state[torchbearer.STOP_TRAINING]:
                    break

                final_metrics.update(self._validation_pass(state))
                state[torchbearer.METRICS] = final_metrics
                state[torchbearer.CALLBACK_LIST].on_end_epoch(state)
                steps_summary = {str(torchbearer.TRAIN_STEPS): state[torchbearer.TRAIN_STEPS], str(torchbearer.VALIDATION_STEPS): state[torchbearer.VALIDATION_STEPS]}
                self.state[torchbearer.HISTORY].append(dict(state[torchbearer.METRICS], **steps_summary))
                state[torchbearer.CALLBACK_LIST].on_checkpoint(state)

                if state[torchbearer.STOP_TRAINING]:
                    break

            state[torchbearer.CALLBACK_LIST].on_end(state)

        return self.state[torchbearer.HISTORY]

    @staticmethod
    def _new_iter(generator):
        if generator is None:
            return None
        if hasattr(generator, 'inf') and generator.inf:  # Inf train loader deals with the iterator itself
            return generator.tb_iter
        else:
            return iter(generator)

    @inject_sampler(torchbearer.TRAIN_DATA, load_batch_standard)
    def _fit_pass(self, state):
        state.update(self.state)  # TODO: Hack to make injection work, should be removed if `self.state` is mutable
        self.train()

        state[torchbearer.ITERATOR] = Trial._new_iter(state[torchbearer.GENERATOR])

        state[torchbearer.METRIC_LIST].reset(state)
        state[torchbearer.METRICS] = {}

        state[torchbearer.STEPS] = 0 if state[torchbearer.STEPS] is None else state[torchbearer.STEPS]
        state[torchbearer.CALLBACK_LIST].on_start_training(state)
        for state[torchbearer.BATCH] in (range(state[torchbearer.STEPS]) if state[torchbearer.STEPS] != -1 else itertools.count()):
            state[torchbearer.SAMPLER](state)
            state[torchbearer.CALLBACK_LIST].on_sample(state)

            # Update parameters
            state[torchbearer.OPTIMIZER].step(lambda: self.closure(state))

            state[torchbearer.METRICS] = state[torchbearer.METRIC_LIST].process(state.data)
            state[torchbearer.CALLBACK_LIST].on_step_training(state)

            if state[torchbearer.STOP_TRAINING]:
                break

        state[torchbearer.METRICS].update(state[torchbearer.METRIC_LIST].process_final(state.data))

        state[torchbearer.CALLBACK_LIST].on_end_training(state)
        return state

    def _test_pass(self, state):
        with torch.no_grad():
            state[torchbearer.ITERATOR] = Trial._new_iter(state[torchbearer.GENERATOR])

            state[torchbearer.METRIC_LIST].reset(state)
            state[torchbearer.METRICS] = {}

            state[torchbearer.CALLBACK_LIST].on_start_validation(state)

            state[torchbearer.STEPS] = 0 if state[torchbearer.STEPS] is None else state[torchbearer.STEPS]
            for state[torchbearer.BATCH] in range(state[torchbearer.STEPS]):
                state[torchbearer.SAMPLER](state)
                state[torchbearer.CALLBACK_LIST].on_sample_validation(state)

                _forward_with_exceptions(torchbearer.X, torchbearer.MODEL, torchbearer.Y_PRED, state)

                state[torchbearer.CALLBACK_LIST].on_forward_validation(state)

                # Loss and metrics
                if torchbearer.Y_TRUE in state:
                    # Loss Calculation
                    try:
                        state[torchbearer.LOSS] = state[torchbearer.CRITERION](state)
                    except TypeError:
                        loss_function_params = _get_param_list(state[torchbearer.Y_PRED]) + _get_param_list(
                            state[torchbearer.Y_TRUE])
                        state[torchbearer.LOSS] = state[torchbearer.CRITERION](*loss_function_params)
                    state[torchbearer.CALLBACK_LIST].on_criterion_validation(state)
                    state[torchbearer.METRICS] = state[torchbearer.METRIC_LIST].process(state.data)

                state[torchbearer.CALLBACK_LIST].on_step_validation(state)
                if state[torchbearer.STOP_TRAINING]:
                    break

            if torchbearer.Y_TRUE in state:
                state[torchbearer.METRICS].update(state[torchbearer.METRIC_LIST].process_final(state.data))
            state[torchbearer.CALLBACK_LIST].on_end_validation(state)
        return state

    @inject_sampler(torchbearer.VALIDATION_DATA, load_batch_standard)
    def _validation_pass(self, state):
        state.update(self.state)  # TODO: Hack to make injection work, should be removed if `self.state` is mutable

        if state[torchbearer.VALIDATION_GENERATOR] is not None or state[torchbearer.VALIDATION_STEPS] is not None:
            self.eval()

            self._test_pass(state)
        return state[torchbearer.METRICS]

    @inject_sampler(torchbearer.VALIDATION_DATA, load_batch_standard)
    @inject_printer(validation_label_letter='e')
    def evaluate(self, verbose=-1, data_key=None):  # Note: kwargs appear unused but are inspected in inject_sampler
        """Evaluate this trial on the validation data.

        Example: ::

            # Simple trial to evaluate on both validation and test data
            >>> from torchbearer import Trial
            >>> test_data = torch.rand(5, 5)
            >>> val_data = torch.rand(5, 5)
            >>> t = Trial(None).with_val_data(val_data).with_test_data(test_data)
            >>> t.evaluate(data_key=torchbearer.VALIDATION_DATA).evaluate(data_key=torchbearer.TEST_DATA)

        Args:
            verbose (int): If 2: use tqdm on batch, If 1: use tqdm on epoch, If 0: display no training progress, If -1: Automatic
            data_key (StateKey): Optional :class:`.StateKey` for the data to evaluate on. Default: torchbearer.VALIDATION_DATA

        Returns:
            dict: The final metric values
        """
        state = State()
        state.update({
            torchbearer.MAX_EPOCHS: 1,
            torchbearer.EPOCH: 0,
            torchbearer.STOP_TRAINING: False
        })
        state.update(self.state)  # TODO: Hack to make injection work, should be removed if `self.state` is mutable

        if state[torchbearer.GENERATOR] is not None or state[torchbearer.STEPS] is not None:
            state[torchbearer.CALLBACK_LIST].on_start(state)
            state[torchbearer.CALLBACK_LIST].on_start_epoch(state)

            self.eval()
            state = self._test_pass(state)

            state[torchbearer.CALLBACK_LIST].on_end_epoch(state)

            if len(self.state[torchbearer.HISTORY]) != 0:
                self.state[torchbearer.HISTORY][-1].update(state[torchbearer.METRICS])

            state[torchbearer.CALLBACK_LIST].on_end(state)
            return state[torchbearer.METRICS]
        return {}

    @inject_callback(AggregatePredictions())
    @inject_sampler(torchbearer.TEST_DATA, load_batch_predict)
    @inject_printer(validation_label_letter='p')
    def predict(self, verbose=-1, data_key=None):  # Note: kwargs appear unused but are inspected in inject_sampler
        """Determine predictions for this trial on the test data.

        Example: ::

            # Simple trial to predict on some validation and test data
            >>> from torchbearer import Trial
            >>> val_data = torch.rand(5, 5)
            >>> test_data = torch.rand(5, 5)
            >>> t = Trial(None).with_test_data(test_data)
            >>> test_predictions = t.predict(data_key=torchbearer.TEST_DATA)

        Args:
            verbose (int): If 2: use tqdm on batch, If 1: use tqdm on epoch, If 0: display no training progress, If -1: Automatic
            data_key (StateKey): Optional :class:`.StateKey` for the data to predict on. Default: torchbearer.TEST_DATA

        Returns:
            list: Model outputs as a list
        """
        state = State()
        state.update({
            torchbearer.MAX_EPOCHS: 1,
            torchbearer.EPOCH: 0,
            torchbearer.STOP_TRAINING: False
        })

        state.update(self.state)  # TODO: Hack to make injection work, should be removed if `self.state` is mutable

        if state[torchbearer.GENERATOR] is not None or state[torchbearer.STEPS] is not None:
            state[torchbearer.CALLBACK_LIST].on_start(state)
            state[torchbearer.CALLBACK_LIST].on_start_epoch(state)

            self.eval()
            res = self._test_pass(state)[torchbearer.FINAL_PREDICTIONS]

            state[torchbearer.CALLBACK_LIST].on_end_epoch(state)
            state[torchbearer.CALLBACK_LIST].on_end(state)
            return res
        return []

    def replay(self, callbacks=None, verbose=2, one_batch=False):  # TODO: Should we track if testing passes have happened?
        """ Replay the fit passes stored in history with given callbacks, useful when reloading a saved Trial. Note that only progress and metric information is populated in state during a replay.

        Example: ::

            >>> from torchbearer import Trial
            >>> state = torch.load('some_state.pt')
            >>> t = Trial(None).load_state_dict(state)
            >>> t.replay()

        Args:
            callbacks (list): List of callbacks to be run during the replay
            verbose (int): If 2: use tqdm on batch, If 1: use tqdm on epoch, If 0: display no training progress
            one_batch (bool): If True, only one batch per epoch is replayed. If False, all batches are replayed

        Returns:
            Trial: self
        """
        if callbacks is None:
            callbacks = []
        history = self.state[torchbearer.HISTORY]
        callbacks.append(get_printer(verbose=verbose, validation_label_letter='v'))
        callbacks = CallbackList(callbacks)

        state = State()
        state.update(self.state)
        state[torchbearer.STOP_TRAINING] = False
        state[torchbearer.MAX_EPOCHS] = len(history)

        callbacks.on_start(state)
        for i in range(len(history)):
            metrics = dict(history[i])
            state[torchbearer.EPOCH] = i
            if not one_batch:
                state[torchbearer.TRAIN_STEPS], state[torchbearer.VALIDATION_STEPS] = metrics[str(torchbearer.TRAIN_STEPS)], metrics[str(torchbearer.VALIDATION_STEPS)]
            else:
                state[torchbearer.TRAIN_STEPS], state[torchbearer.VALIDATION_STEPS] =\
                    1 if metrics[str(torchbearer.TRAIN_STEPS)] is not None else None,\
                    1 if metrics[str(torchbearer.VALIDATION_STEPS)] is not None else None
            del metrics[str(torchbearer.TRAIN_STEPS)]
            del metrics[str(torchbearer.VALIDATION_STEPS)]
            state[torchbearer.METRICS] = metrics

            self._replay_pass(state, callbacks)
        callbacks.on_end(state)

        return self

    def _replay_pass(self, state, callback_list):
        callback_list.on_start_epoch(state)
        all_metrics = state[torchbearer.METRICS]

        if state[torchbearer.TRAIN_STEPS] is not None:
            # Training pass
            state[torchbearer.STEPS] = state[torchbearer.TRAIN_STEPS]
            state[torchbearer.METRICS] = {key: all_metrics[key] for key in all_metrics.keys() if "val_" not in key}
            callback_list.on_start_training(state)
            for state[torchbearer.BATCH] in range(state[torchbearer.STEPS]):
                callback_list.on_sample(state)
                callback_list.on_forward(state)
                callback_list.on_criterion(state)
                callback_list.on_backward(state)
                callback_list.on_step_training(state)
                if state[torchbearer.STOP_TRAINING]:
                    break
            callback_list.on_end_training(state)

        if state[torchbearer.VALIDATION_STEPS] is not None:
            # Validation pass
            if not state[torchbearer.STOP_TRAINING]:
                state[torchbearer.STEPS] = state[torchbearer.VALIDATION_STEPS]
                state[torchbearer.METRICS] = {key: all_metrics[key] for key in all_metrics.keys() if "val_" in key}
                callback_list.on_start_validation(state)
                for state[torchbearer.BATCH] in range(state[torchbearer.STEPS]):
                    callback_list.on_sample_validation(state)
                    callback_list.on_forward_validation(state)
                    callback_list.on_criterion_validation(state)
                    callback_list.on_step_validation(state)
                    if state[torchbearer.STOP_TRAINING]:
                        break
                callback_list.on_end_validation(state)

        state[torchbearer.METRICS] = all_metrics
        callback_list.on_end_epoch(state)

        return self

    # Device management

    def train(self):
        """Set model and metrics to training mode.

        Example: ::
            >>> from torchbearer import Trial
            >>> t = Trial(None).train()

        Returns:
            Trial: self
        """
        self.state[torchbearer.MODEL].train()
        self.state[torchbearer.METRIC_LIST].train()

        return self

    def eval(self):
        """Set model and metrics to evaluation mode

        Example: ::
            >>> from torchbearer import Trial
            >>> t = Trial(None).eval()

        Returns:
            Trial: self
        """
        self.state[torchbearer.MODEL].eval()
        if torchbearer.DATA in self.state:
            self.state[torchbearer.METRIC_LIST].eval(data_key=self.state[torchbearer.DATA])
        else:
            self.state[torchbearer.METRIC_LIST].eval()

        return self

    def to(self, *args, **kwargs):
        """ Moves and/or casts the parameters and buffers.

        Example: ::
            >>> from torchbearer import Trial
            >>> t = Trial(None).to('cuda:1')

        Args:
            args: See: `torch.nn.Module.to <https://pytorch.org/docs/stable/nn.html?highlight=#torch.nn.Module.to>`_
            kwargs: See: `torch.nn.Module.to <https://pytorch.org/docs/stable/nn.html?highlight=#torch.nn.Module.to>`_

        Returns:
            Trial: self
        """
        self.state[torchbearer.MODEL].to(*args, **kwargs)

        for state in self.state[torchbearer.OPTIMIZER].state.values():
            for k, v in state.items():
                if torch.is_tensor(v):
                    state[k] = v.to(*args, **kwargs)

        self.state = update_device_and_dtype(self.state, *args, **kwargs)

        return self

    def cuda(self, device=None):
        """ Moves all model parameters and buffers to the GPU.

        Example: ::
            >>> from torchbearer import Trial
            >>> t = Trial(None).cuda()

        Args:
            device (int): if specified, all parameters will be copied to that device

        Returns:
            Trial: self
        """
        if device is None:
            device = torch.cuda.current_device()
        self.to('cuda:' + str(device))

        return self

    def cpu(self):
        """ Moves all model parameters and buffers to the CPU.

        Example: ::
            >>> from torchbearer import Trial
            >>> t = Trial(None).cpu()

        Returns:
            Trial: self
        """
        self.to('cpu')

        return self

    # States

    def state_dict(self, **kwargs):
        """Get a dict containing the model and optimizer states, as well as the model history.

        Example: ::
            >>> from torchbearer import Trial
            >>> t = Trial(None)
            >>> state = t.state_dict() # State dict that can now be saved with torch.save

        Args:
            kwargs: See: `torch.nn.Module.state_dict <https://pytorch.org/docs/stable/nn.html?highlight=#torch.nn.Module.state_dict>`_

        Returns:
            dict: A dict containing parameters and persistent buffers.
        """
        state_dict = {
            torchbearer.VERSION: torchbearer.__version__.replace('.dev', ''),
            torchbearer.MODEL: self.state[torchbearer.MODEL].state_dict(**kwargs),
            torchbearer.OPTIMIZER: self.state[torchbearer.OPTIMIZER].state_dict(),
            torchbearer.HISTORY: self.state[torchbearer.HISTORY],
            torchbearer.CALLBACK_LIST: self.state[torchbearer.CALLBACK_LIST].state_dict()
        }
        return state_dict

    def load_state_dict(self, state_dict, resume=True, **kwargs):
        """Resume this trial from the given state. Expects that this trial was constructed in the same way. Optionally,
        just load the model state when resume=False.

        Example: ::
            >>> from torchbearer import Trial
            >>> t = Trial(None)
            >>> state = torch.load('some_state.pt')
            >>> t.load_state_dict(state)

        Args:
            state_dict (dict): The state dict to reload
            resume (bool): If True, resume from the given state. Else, just load in the model weights.
            kwargs: See: `torch.nn.Module.load_state_dict <https://pytorch.org/docs/stable/nn.html?highlight=#torch.nn.Module.load_state_dict>`_

        Returns:
            Trial: self
        """
        if resume and torchbearer.MODEL in state_dict:  # torchbearer dict
            if torchbearer.VERSION in state_dict and state_dict[torchbearer.VERSION] != torchbearer.__version__.replace('.dev', ''):
                warnings.warn('This state dict was saved with a different torchbearer version, loading available keys. Consider setting resume=False')

            if torchbearer.MODEL in state_dict:
                self.state[torchbearer.MODEL].load_state_dict(state_dict[torchbearer.MODEL], **kwargs)

            if torchbearer.OPTIMIZER in state_dict:
                self.state[torchbearer.OPTIMIZER].load_state_dict(state_dict[torchbearer.OPTIMIZER])

            if torchbearer.HISTORY in state_dict:
                self.state[torchbearer.HISTORY] = state_dict[torchbearer.HISTORY]

            if torchbearer.CALLBACK_LIST in state_dict:
                self.state[torchbearer.CALLBACK_LIST].load_state_dict(state_dict[torchbearer.CALLBACK_LIST])
        elif torchbearer.MODEL in state_dict:
            self.state[torchbearer.MODEL].load_state_dict(state_dict[torchbearer.MODEL], **kwargs)
        else:  # something else
            warnings.warn('Not a torchbearer state dict, passing to model')
            self.state[torchbearer.MODEL].load_state_dict(state_dict, **kwargs)

        return self



import unittest


class TestMagics(unittest.TestCase):
    def test_torchbearer_fn(self):
        from torchbearer.magics import torchbearer as tb
        from torchbearer.magics import is_notebook

        self.assertFalse(is_notebook())
        tb('notebook')
        self.assertTrue(is_notebook())
        tb('normal')
        self.assertFalse(is_notebook())




from unittest import TestCase

from mock import Mock, MagicMock, patch

import torchbearer
from torchbearer.metrics.timer import TimerMetric, _TimerMetric


class TestTimer(TestCase):
    def test_update_time(self):
        timer = TimerMetric('test')
        timerMetric = TimerMetric('test2')
        timerMetric.process = Mock(return_value=1)
        timer.update_time('test', timerMetric, {})
        self.assertTrue(timer.get_timings()['test'] == 1)

        timerMetric.process = Mock(return_value=2)
        timer.update_time('test_2', timerMetric, {})
        self.assertTrue(timer.get_timings()['test_2'] == 2)

    def test_calls(self):
        timer = TimerMetric('test')
        timer.batch_timer = MagicMock()
        timer.epoch_timer = MagicMock()
        timer.train_timer = MagicMock()
        timer.total_timer = MagicMock()
        timer.valid_timer = MagicMock()

        timer.on_start({})
        timer.on_start_training({})
        timer.on_start_epoch({})
        timer.on_sample({})
        timer.on_forward({})
        timer.on_criterion({})
        timer.on_backward({})
        timer.on_step_training({})
        timer.on_start_validation({})
        timer.on_sample_validation({})
        timer.on_forward_validation({})
        timer.on_criterion_validation({})
        timer.on_step_validation({})
        timer.on_end_training({})
        timer.on_end_validation({})
        timer.on_end_epoch({})
        timer.on_end({})

        self.assertTrue(timer.batch_timer.process.call_count == 11)
        self.assertTrue(timer.total_timer.process.call_count == 1)
        self.assertTrue(timer.epoch_timer.process.call_count == 1)
        self.assertTrue(timer.train_timer.process.call_count == 2)
        self.assertTrue(timer.valid_timer.process.call_count == 1)

    def test_process(self):
        timer = TimerMetric((torchbearer.metrics.timer.ON_FORWARD, ))
        timer.time_dict = {torchbearer.metrics.timer.ON_FORWARD: 1, 'test': 2}
        self.assertTrue(timer.process({})[torchbearer.metrics.timer.ON_FORWARD] == 1)

    def test_reset(self):
        state = {torchbearer.CALLBACK_LIST: torchbearer.callbacks.CallbackList([])}

        timer = TimerMetric()
        self.assertTrue(state[torchbearer.CALLBACK_LIST].callback_list == [])
        timer.reset(state)
        self.assertIsInstance(state[torchbearer.CALLBACK_LIST].callback_list[0], TimerMetric)

        timer.reset(state)
        self.assertTrue(len(state[torchbearer.CALLBACK_LIST].callback_list) == 1)


class TestTimerMetric(TestCase):
    @patch('time.time')
    def test_process(self, time):
        time.return_value = 1
        timer_metric = _TimerMetric('test')
        time.return_value = 2
        dt = timer_metric.process({})

        self.assertTrue(dt == 1)

    @patch('time.time')
    def test_reset(self, time):
        time.return_value = 1
        timer_metric = _TimerMetric('test')

        time.return_value = 3
        timer_metric.reset({})
        self.assertTrue(timer_metric.t == 3)




import torchbearer
from torchbearer import Callback
import torch
from torch.distributions import Beta

from torchbearer.bases import cite

cutout = """
@article{devries2017improved,
  title={Improved regularization of convolutional neural networks with Cutout},
  author={DeVries, Terrance and Taylor, Graham W},
  journal={arXiv preprint arXiv:1708.04552},
  year={2017}
}
"""

random_erase = """
@article{zhong2017random,
  title={Random erasing data augmentation},
  author={Zhong, Zhun and Zheng, Liang and Kang, Guoliang and Li, Shaozi and Yang, Yi},
  journal={arXiv preprint arXiv:1708.04896},
  year={2017}
}
"""


cutmix = """
@article{yun2019cutmix,
  title={Cutmix: Regularization strategy to train strong classifiers with localizable features},
  author={Yun, Sangdoo and Han, Dongyoon and Oh, Seong Joon and Chun, Sanghyuk and Choe, Junsuk and Yoo, Youngjoon},
  journal={arXiv preprint arXiv:1905.04899},
  year={2019}
}
"""


@cite(cutout)
class Cutout(Callback):
    """ Cutout callback which randomly masks out patches of image data. Implementation a modified version of the code
    found `here <https://github.com/uoguelph-mlrg/Cutout>`_.

    Example: ::

        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import Cutout

        # Example Trial which does Cutout regularisation
        >>> cutout = Cutout(1, 10)
        >>> trial = Trial(None, callbacks=[cutout], metrics=['acc'])

    Args:
        n_holes (int): Number of patches to cut out of each image.
        length (int): The length (in pixels) of each square patch.
        constant (float): Constant value for each square patch

    State Requirements:
        - :attr:`torchbearer.state.X`: State should have the current data stored
    """
    def __init__(self, n_holes, length, constant=0.):
        super(Cutout, self).__init__()
        self.constant = constant
        self.cutter = BatchCutout(n_holes, length, length)

    def on_sample(self, state):
        super(Cutout, self).on_sample(state)
        mask = self.cutter(state[torchbearer.X])
        erase_locations = mask == 0
        constant = torch.ones_like(state[torchbearer.X]) * self.constant
        state[torchbearer.X][erase_locations] = constant[erase_locations]


@cite(random_erase)
class RandomErase(Callback):
    """ Random erase callback which replaces random patches of image data with random noise.
    Implementation a modified version of the cutout code found
    `here <https://github.com/uoguelph-mlrg/Cutout>`_.

    Example: ::

        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import RandomErase

        # Example Trial which does Cutout regularisation
        >>> erase = RandomErase(1, 10)
        >>> trial = Trial(None, callbacks=[erase], metrics=['acc'])

    Args:
        n_holes (int): Number of patches to cut out of each image.
        length (int): The length (in pixels) of each square patch.

    State Requirements:
        - :attr:`torchbearer.state.X`: State should have the current data stored
    """
    def __init__(self, n_holes, length):
        super(RandomErase, self).__init__()
        self.cutter = BatchCutout(n_holes, length, length)

    def on_sample(self, state):
        super(RandomErase, self).on_sample(state)
        mask = self.cutter(state[torchbearer.X])
        erase_locations = mask == 0
        random = torch.rand_like(state[torchbearer.X])
        state[torchbearer.X][erase_locations] = random[erase_locations]


@cite(cutmix)
class CutMix(Callback):
    """ Cutmix callback which replaces a random patch of image data with the corresponding patch from another image.
    This callback also converts labels to one hot before combining them according to the lambda parameters, sampled from
    a beta distribution as is done in the paper.

    Example: ::

        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import CutMix

        # Example Trial which does CutMix regularisation
        >>> cutmix = CutMix(1, classes=10)
        >>> trial = Trial(None, callbacks=[cutmix], metrics=['acc'])

    Args:
        alpha (float): The alpha value for the beta distribution.
        classes (int): The number of classes for conversion to one hot.

    State Requirements:
        - :attr:`torchbearer.state.X`: State should have the current data stored
        - :attr:`torchbearer.state.Y_TRUE`: State should have the current data stored
    """
    def __init__(self, alpha, classes=-1):
        super(CutMix, self).__init__()
        self.classes = classes
        self.dist = Beta(torch.tensor([float(alpha)]), torch.tensor([float(alpha)]))

    def _to_one_hot(self, target):
        if target.dim() == 1:
            target = target.unsqueeze(1)
            one_hot = torch.zeros_like(target).repeat(1, self.classes)
            one_hot.scatter_(1, target, 1)
            return one_hot
        return target

    def on_sample(self, state):
        super(CutMix, self).on_sample(state)

        lam = self.dist.sample().to(state[torchbearer.DEVICE])
        length = (1 - lam).sqrt()
        cutter = BatchCutout(1, (length * state[torchbearer.X].size(-1)).round().item(), (length * state[torchbearer.X].size(-2)).round().item())
        mask = cutter(state[torchbearer.X])
        erase_locations = mask == 0

        permutation = torch.randperm(state[torchbearer.X].size(0))

        state[torchbearer.X][erase_locations] = state[torchbearer.X][permutation][erase_locations]

        target = self._to_one_hot(state[torchbearer.TARGET]).float()
        state[torchbearer.TARGET] = lam * target + (1 - lam) * target[permutation]

    def on_sample_validation(self, state):
        super(CutMix, self).on_sample_validation(state)
        state[torchbearer.TARGET] = self._to_one_hot(state[torchbearer.TARGET]).float()


class BatchCutout(object):
    """Randomly mask out one or more patches from a batch of images.

    Args:
        n_holes (int): Number of patches to cut out of each image.
        width (int): The width (in pixels) of each square patch.
        height (int): The height (in pixels) of each square patch.
    """
    def __init__(self, n_holes, width, height):
        self.n_holes = n_holes
        self.width = width
        self.height = height

    def __call__(self, img):
        """

        Args:
            img (Tensor): Tensor image of size (B, C, H, W).
        Returns:
            Tensor: Image with n_holes of dimension length x length cut out of it.
        """
        b = img.size(0)
        c = img.size(1)
        h = img.size(-2)
        w = img.size(-1)

        mask = torch.ones((b, h, w), device=img.device)

        for n in range(self.n_holes):
            y = torch.randint(h, (b,)).long()
            x = torch.randint(w, (b,)).long()

            y1 = (y - self.height // 2).clamp(0, h).int()
            y2 = (y + self.height // 2).clamp(0, h).int()
            x1 = (x - self.width // 2).clamp(0, w).int()
            x2 = (x + self.width // 2).clamp(0, w).int()

            for batch in range(b):
                mask[batch, y1[batch]: y2[batch], x1[batch]: x2[batch]] = 0

        mask = mask.unsqueeze(1).repeat(1, c, 1, 1)

        return mask



import torch
import torch.nn.functional as F
from torch.distributions.beta import Beta

import torchbearer
from torchbearer import cite
from torchbearer.callbacks import Callback
from torchbearer.metrics import CategoricalAccuracy, AdvancedMetric, running_mean, mean, super


mixup= """
@inproceedings{zhang2018mixup,
  title={mixup: Beyond Empirical Risk Minimization},
  author={Hongyi Zhang and Moustapha Cisse and Yann N. Dauphin and David Lopez-Paz},
  booktitle={International Conference on Learning Representations},
  year={2018}
}
"""


@running_mean
@mean
class MixupAcc(AdvancedMetric):
    def __init__(self):
        super(MixupAcc, self).__init__('mixup_acc')
        self.cat_acc = CategoricalAccuracy().root

    def process_train(self, *args):
        super(MixupAcc, self).process_train(*args)
        state = args[0]
        
        target1, target2 = state[torchbearer.Y_TRUE]
        _state = args[0].copy()
        _state[torchbearer.Y_TRUE] = target1
        acc1 = self.cat_acc.process(_state)

        _state = args[0].copy()
        _state[torchbearer.Y_TRUE] = target2
        acc2 = self.cat_acc.process(_state)

        return acc1 * state[torchbearer.MIXUP_LAMBDA] + acc2 * (1-state[torchbearer.MIXUP_LAMBDA])

    def process_validate(self, *args):
        super(MixupAcc, self).process_validate(*args)

        return self.cat_acc.process(*args)

    def reset(self, state):
        self.cat_acc.reset(state)


@cite(mixup)
class Mixup(Callback):
    """Perform mixup on the model inputs. Requires use of :meth:`MixupInputs.loss`, otherwise lambdas can be found in
    state under :attr:`.MIXUP_LAMBDA`. Model targets will be a tuple containing the original target and permuted target.

    .. note::

        The accuracy metric for mixup is different on training to deal with the different targets,
        but for validation it is exactly the categorical accuracy, despite being called "val_mixup_acc"

    Example: ::

        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import Mixup

        # Example Trial which does Mixup regularisation
        >>> mixup = Mixup(0.9)
        >>> trial = Trial(None, criterion=Mixup.mixup_loss, callbacks=[mixup], metrics=['acc'])

    Args:
        lam (float): Mixup inputs by fraction lam. If RANDOM, choose lambda from Beta(alpha, alpha). Else, lambda=lam
        alpha (float): The alpha value to use in the beta distribution.
    """
    RANDOM = -10.0

    def __init__(self, alpha=1.0, lam=RANDOM):
        super(Mixup, self).__init__()
        self.alpha = alpha
        self.lam = lam
        self.distrib = Beta(self.alpha, self.alpha)

    @staticmethod
    def mixup_loss(state):
        """The standard cross entropy loss formulated for mixup (weighted combination of `F.cross_entropy`).

        Args:
            state: The current :class:`Trial` state.
        """
        input, target = state[torchbearer.Y_PRED], state[torchbearer.Y_TRUE]

        if state[torchbearer.DATA] is torchbearer.TRAIN_DATA:
            y1, y2 = target
            return F.cross_entropy(input, y1) * state[torchbearer.MIXUP_LAMBDA] + F.cross_entropy(input, y2) * (1-state[torchbearer.MIXUP_LAMBDA])
        else:
            return F.cross_entropy(input, target)

    def on_sample(self, state):
        if self.lam is Mixup.RANDOM:
            if self.alpha > 0:
                lam = self.distrib.sample()
            else:
                lam = 1.0
        else:
            lam = self.lam

        state[torchbearer.MIXUP_LAMBDA] = lam

        state[torchbearer.MIXUP_PERMUTATION] = torch.randperm(state[torchbearer.X].size(0))
        state[torchbearer.X] = state[torchbearer.X] * state[torchbearer.MIXUP_LAMBDA] + state[torchbearer.X][state[torchbearer.MIXUP_PERMUTATION], :] * (1-state[torchbearer.MIXUP_LAMBDA])
        state[torchbearer.Y_TRUE] = (state[torchbearer.Y_TRUE], state[torchbearer.Y_TRUE][state[torchbearer.MIXUP_PERMUTATION]])


from torchbearer.metrics import default as d
d.__loss_map__[Mixup.mixup_loss.__name__] = MixupAcc



import unittest

from mock import Mock

import torchbearer.metrics as metrics
from torchbearer.metrics import default_for_key, lambda_metric, EpochLambda, BatchLambda


class TestDecorators(unittest.TestCase):
    def test_default_for_key_class(self):
        metric = metrics.Loss
        metric = default_for_key('test')(metric)
        self.assertTrue(metrics.get_default('test').name == 'loss')
        self.assertTrue(metric == metrics.Loss)

    def test_default_for_key_metric(self):
        metric = metrics.Loss()
        metric = default_for_key('test')(metric)
        self.assertTrue(metrics.get_default('test').name == 'loss')
        self.assertTrue(metric.name == 'loss')

    def test_default_for_key_args(self):
        mock = Mock()

        class MyMetric(metrics.Metric):
            def __init__(self, *args, **kwargs):
                super(MyMetric, self).__init__('test')
                mock(*args, **kwargs)

        default_for_key('test', 10, some_arg='a test')(MyMetric)
        metrics.get_default('test')
        mock.assert_called_once_with(10, some_arg='a test')

    def test_lambda_metric_epoch(self):
        metric = 'test'
        metric = lambda_metric('test', on_epoch=True)(metric)
        self.assertTrue(isinstance(metric, EpochLambda))
        self.assertTrue(metric._final == 'test')

    def test_lambda_metric_batch(self):
        metric = 'test'
        metric = lambda_metric('test')(metric)
        self.assertTrue(isinstance(metric, BatchLambda))
        self.assertTrue(metric._metric_function == 'test')

    def test_to_dict_metric_class(self):
        metric = metrics.Metric
        out = metrics.to_dict(metric)('test')
        self.assertTrue(out.metric.name == 'test')
        self.assertTrue(isinstance(out, metrics.ToDict))

    def test_to_dict_metric_instance(self):
        metric = metrics.Metric('test')
        out = metrics.to_dict(metric)
        self.assertTrue(out.metric.name == 'test')
        self.assertTrue(isinstance(out, metrics.ToDict))

    def test_mean_metric_class(self):
        metric = metrics.Metric
        out = metrics.mean(dim=10)(metric)('test')
        self.assertTrue(isinstance(out, metrics.MetricTree))
        self.assertTrue(isinstance(out.children[0], metrics.ToDict))
        self.assertTrue(isinstance(out.children[0].metric, metrics.Mean))
        self.assertTrue(out.children[0].metric._kwargs['dim'] == 10)
        self.assertTrue(out.children[0].metric.name == 'test')
        self.assertTrue(out.root.name == 'test')

    def test_mean_metric_instance(self):
        metric = metrics.Metric('test')
        out = metrics.mean(metric, dim=10)
        self.assertTrue(isinstance(out, metrics.MetricTree))
        self.assertTrue(isinstance(out.children[0], metrics.ToDict))
        self.assertTrue(isinstance(out.children[0].metric, metrics.Mean))
        self.assertTrue(out.children[0].metric._kwargs['dim'] == 10)
        self.assertTrue(out.children[0].metric.name == 'test')
        self.assertTrue(out.root.name == 'test')

    def test_std_metric_class(self):
        metric = metrics.Metric
        out = metrics.std(dim=10, unbiased=False)(metric)('test')
        self.assertTrue(isinstance(out, metrics.MetricTree))
        self.assertTrue(isinstance(out.children[0], metrics.ToDict))
        self.assertTrue(isinstance(out.children[0].metric, metrics.Std))
        self.assertTrue(out.children[0].metric._kwargs['dim'] == 10)
        self.assertTrue(not out.children[0].metric._unbiased)
        self.assertTrue(out.children[0].metric.name == 'test_std')
        self.assertTrue(out.root.name == 'test')

    def test_std_metric_instance(self):
        metric = metrics.Metric('test')
        out = metrics.std(metric, dim=10, unbiased=False)
        self.assertTrue(isinstance(out, metrics.MetricTree))
        self.assertTrue(isinstance(out.children[0], metrics.ToDict))
        self.assertTrue(isinstance(out.children[0].metric, metrics.Std))
        self.assertTrue(out.children[0].metric._kwargs['dim'] == 10)
        self.assertTrue(not out.children[0].metric._unbiased)
        self.assertTrue(out.children[0].metric.name == 'test_std')
        self.assertTrue(out.root.name == 'test')

    def test_var_metric_class(self):
        metric = metrics.Metric
        out = metrics.var(dim=10, unbiased=False)(metric)('test')
        self.assertTrue(isinstance(out, metrics.MetricTree))
        self.assertTrue(isinstance(out.children[0], metrics.ToDict))
        self.assertTrue(isinstance(out.children[0].metric, metrics.Var))
        self.assertTrue(out.children[0].metric._kwargs['dim'] == 10)
        self.assertTrue(not out.children[0].metric._unbiased)
        self.assertTrue(out.children[0].metric.name == 'test_var')
        self.assertTrue(out.root.name == 'test')

    def test_var_metric_instance(self):
        metric = metrics.Metric('test')
        out = metrics.var(metric, dim=10, unbiased=False)
        self.assertTrue(isinstance(out, metrics.MetricTree))
        self.assertTrue(isinstance(out.children[0], metrics.ToDict))
        self.assertTrue(isinstance(out.children[0].metric, metrics.Var))
        self.assertTrue(out.children[0].metric._kwargs['dim'] == 10)
        self.assertTrue(not out.children[0].metric._unbiased)
        self.assertTrue(out.children[0].metric.name == 'test_var')
        self.assertTrue(out.root.name == 'test')

    def test_running_mean_metric_class(self):
        metric = metrics.Metric
        out = metrics.running_mean(batch_size=40, step_size=20, dim=10)(metric)('test')
        self.assertTrue(isinstance(out, metrics.MetricTree))
        self.assertTrue(isinstance(out.children[0], metrics.ToDict))
        self.assertTrue(isinstance(out.children[0].metric, metrics.RunningMean))
        self.assertTrue(out.children[0].metric._batch_size == 40)
        self.assertTrue(out.children[0].metric._step_size == 20)
        self.assertTrue(out.children[0].metric._kwargs['dim'] == 10)
        self.assertTrue(out.children[0].metric.name == 'running_test')
        self.assertTrue(out.root.name == 'test')

    def test_running_mean_metric_instance(self):
        metric = metrics.Metric('test')
        out = metrics.running_mean(batch_size=40, step_size=20, dim=10)(metric)
        self.assertTrue(isinstance(out, metrics.MetricTree))
        self.assertTrue(isinstance(out.children[0], metrics.ToDict))
        self.assertTrue(isinstance(out.children[0].metric, metrics.RunningMean))
        self.assertTrue(out.children[0].metric._batch_size == 40)
        self.assertTrue(out.children[0].metric._step_size == 20)
        self.assertTrue(out.children[0].metric._kwargs['dim'] == 10)
        self.assertTrue(out.children[0].metric.name == 'running_test')
        self.assertTrue(out.root.name == 'test')



import inspect
from torchbearer import Metric, no_grad

__defaults__ = {}


def add_default(key, metric, *args, **kwargs):
    __defaults__[key] = (metric, args, kwargs)


def get_default(key):
    metric, args, kwargs = __defaults__[key]
    if inspect.isclass(metric):
        metric = metric(*args, **kwargs)
    return metric


class MetricTree(Metric):
    """A tree structure which has a node :class:`.Metric` and some children. Upon execution, the node is called with the
    input and its output is passed to each of the children. A dict is updated with the results.

    .. note::

       If the node output is already a dict (i.e. the node is a standalone metric), this is unwrapped before passing the
       **first** value to the children.

    Args:
        metric (Metric): The metric to act as the root node of the tree / subtree
    """
    def __init__(self, metric):
        super(MetricTree, self).__init__(metric.name)
        self.root = metric
        self.children = []

    def __str__(self):
        return str(self.root)

    def add_child(self, child):
        """Add a child to this node of the tree

        Args:
            child (Metric): The child to add
        """
        self.children.append(child)

    def _for_tree(self, function, *args):
        result = {}
        node_value = function(self.root, *args)
        if isinstance(node_value, dict):
            node_value = next(iter(node_value.values()))

        for subtree in self.children:
            out = function(subtree, node_value)
            if out is not None:
                result.update(out)

        return result

    def process(self, *args):
        """Process this node and then pass the output to each child.

        Returns:
            A dict containing all results from the children
        """
        return self._for_tree(lambda metric, *in_args: metric.process(*in_args), *args)

    def process_final(self, *args):
        """Process this node and then pass the output to each child.

        Returns:
            A dict containing all results from the children
        """
        return self._for_tree(lambda metric, *in_args: metric.process_final(*in_args), *args)

    def eval(self, data_key=None):
        self.root.eval(data_key=data_key)

        for subtree in self.children:
            subtree.eval(data_key=data_key)

    def train(self):
        self.root.train()

        for subtree in self.children:
            subtree.train()

    def reset(self, state):
        self.root.reset(state)

        for subtree in self.children:
            subtree.reset(state)


class MetricList(Metric):
    """The :class:`MetricList` class is a wrapper for a list of metrics which acts as a single metric and produces a
    dictionary of outputs.

    Args:
        metric_list (list): The list of metrics to be wrapped. If the list contains a :class:`MetricList`, this will be
            unwrapped. Any strings in the list will be retrieved from metrics.DEFAULT_METRICS.
    """

    def __init__(self, metric_list):
        super(MetricList, self).__init__('metric_list')

        self.metric_list = []

        for metric in metric_list:

            if isinstance(metric, str):
                metric = get_default(metric)

            if isinstance(metric, MetricList):
                self.metric_list = self.metric_list + metric.metric_list
            else:
                self.metric_list.append(metric)

    def _for_list(self, function):
        result = {}
        for metric in self.metric_list:
            out = function(metric)
            if out is not None:
                result.update(out)
        return result

    @no_grad()
    def process(self, *args):
        """Process each metric an wrap in a dictionary which maps metric names to values.

        Returns:
            dict[str,any]: A dictionary which maps metric names to values.
        """
        return self._for_list(lambda metric: metric.process(*args))

    @no_grad()
    def process_final(self, *args):
        """Process each metric an wrap in a dictionary which maps metric names to values.

        Returns:
            dict[str,any]: A dictionary which maps metric names to values.

        """
        return self._for_list(lambda metric: metric.process_final(*args))

    def train(self):
        """Put each metric in train mode.
        """
        self._for_list(lambda metric: metric.train())

    def eval(self, data_key=None):
        """Put each metric in eval mode
        """
        self._for_list(lambda metric: metric.eval(data_key=data_key))

    def reset(self, state):
        """Reset each metric with the given state.

        Args:
            state: The current state dict of the :class:`.Trial`.
        """
        self._for_list(lambda metric: metric.reset(state))

    def __str__(self):
        return str([str(m) for m in self.metric_list])


class AdvancedMetric(Metric):
    """The :class:`AdvancedMetric` class is a metric which provides different process methods for training and
    validation. This enables running metrics which do not output intermediate steps during validation.

    Args:
        name (str): The name of the metric.
    """

    def __init__(self, name):
        super(AdvancedMetric, self).__init__(name)
        self._process = self.process_train
        self._process_final = self.process_final_train

    def process_train(self, *args):
        """Process the given state and return the metric value for a training iteration.

        Returns:
            The metric value for a training iteration.
        """
        pass

    def process_validate(self, *args):
        """Process the given state and return the metric value for a validation iteration.

        Returns:
            The metric value for a validation iteration.
        """
        pass

    def process_final_train(self, *args):
        """Process the given state and return the final metric value for a training iteration.

        Returns:
            The final metric value for a training iteration.
        """
        pass

    def process_final_validate(self, *args):
        """Process the given state and return the final metric value for a validation iteration.

        Returns:
            The final metric value for a validation iteration.
        """
        pass

    def process(self, *args):
        """Depending on the current mode, return the result of either 'process_train' or 'process_validate'.

        Returns:
            The metric value.
        """
        return self._process(*args)

    def process_final(self, *args):
        """Depending on the current mode, return the result of either 'process_final_train' or 'process_final_validate'.

        Returns:
            The final metric value.
        """
        return self._process_final(*args)

    def eval(self, data_key=None):
        """Put the metric in eval mode.

        Args:
            data_key (StateKey): The torchbearer data_key, if used
        """
        self._process = self.process_validate
        self._process_final = self.process_final_validate

    def train(self):
        """Put the metric in train mode.
        """
        self._process = self.process_train
        self._process_final = self.process_final_train



import torchbearer

from torchbearer.callbacks import Callback

import torch


class TorchScheduler(Callback):
    def __init__(self, scheduler_builder, monitor=None, step_on_batch=False):
        self._scheduler_builder = scheduler_builder
        self._monitor = monitor
        self._scheduler = None
        self._step_on_batch = step_on_batch

    def on_start(self, state):
        self._scheduler = self._scheduler_builder(state[torchbearer.OPTIMIZER])

    def on_sample(self, state):
        if self._step_on_batch and self._monitor is None:
            self._scheduler.step()

    def on_step_training(self, state):
        if self._step_on_batch and self._monitor is not None:
            self._scheduler.step(state[torchbearer.METRICS][self._monitor])

    def on_start_training(self, state):
        if not self._step_on_batch and self._monitor is None:
            self._scheduler.step(epoch=state[torchbearer.EPOCH])

    def on_end_epoch(self, state):
        if not self._step_on_batch and self._monitor is not None:
            self._scheduler.step(state[torchbearer.METRICS][self._monitor], epoch=state[torchbearer.EPOCH])


class LambdaLR(TorchScheduler):
    """
    Example: ::

        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import LambdaLR

        # Example Trial which performs the two learning rate lambdas from the PyTorch docs
        >>> lambda1 = lambda epoch: epoch // 30
        >>> lambda2 = lambda epoch: 0.95 ** epoch
        >>> scheduler = LambdaLR(lr_lambda=[lambda1, lambda2])
        >>> trial = Trial(None, callbacks=[scheduler], metrics=['loss'], verbose=2).for_steps(10).run(1)

    Args:
        step_on_batch (bool): If True, step will be called on each training iteration rather than on each epoch

    See:
        `PyTorch LambdaLR <http://pytorch.org/docs/master/optim.html#torch.optim.lr_scheduler.LambdaLR>`_
    """
    def __init__(self, lr_lambda, last_epoch=-1, step_on_batch=False):
        super(LambdaLR, self).__init__(lambda opt:
                                       torch.optim.lr_scheduler.LambdaLR(opt, lr_lambda, last_epoch=last_epoch),
                                       step_on_batch=step_on_batch)


class StepLR(TorchScheduler):
    """
    Example: ::

        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import StepLR

        >>> # Assuming optimizer uses lr = 0.05 for all groups
        >>> # lr = 0.05     if epoch < 30
        >>> # lr = 0.005    if 30 <= epoch < 60
        >>> # lr = 0.0005   if 60 <= epoch < 90
        >>> scheduler = StepLR(step_size=30, gamma=0.1)
        >>> trial = Trial(None, callbacks=[scheduler], metrics=['loss'], verbose=2).for_steps(10).run(1)

    Args:
        step_on_batch (bool): If True, step will be called on each training iteration rather than on each epoch

    See:
        `PyTorch StepLR <http://pytorch.org/docs/master/optim.html#torch.optim.lr_scheduler.StepLR>`_
    """
    def __init__(self, step_size, gamma=0.1, last_epoch=-1, step_on_batch=False):
        super(StepLR, self).__init__(lambda opt:
                                     torch.optim.lr_scheduler.StepLR(opt, step_size, gamma=gamma,
                                                                     last_epoch=last_epoch),
                                     step_on_batch=step_on_batch)


class MultiStepLR(TorchScheduler):
    """
    Example: ::

        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import MultiStepLR

        >>> # Assuming optimizer uses lr = 0.05 for all groups
        >>> # lr = 0.05     if epoch < 30
        >>> # lr = 0.005    if 30 <= epoch < 80
        >>> # lr = 0.0005   if epoch >= 80
        >>> scheduler = MultiStepLR(milestones=[30,80], gamma=0.1)
        >>> trial = Trial(None, callbacks=[scheduler], metrics=['loss'], verbose=2).for_steps(10).run(1)

    Args:
        step_on_batch (bool): If True, step will be called on each training iteration rather than on each epoch

    See:
        `PyTorch MultiStepLR <http://pytorch.org/docs/master/optim.html#torch.optim.lr_scheduler.MultiStepLR>`_
    """
    def __init__(self, milestones, gamma=0.1, last_epoch=-1, step_on_batch=False):
        super(MultiStepLR, self).__init__(lambda opt:
                                          torch.optim.lr_scheduler.MultiStepLR(opt, milestones, gamma=gamma,
                                                                               last_epoch=last_epoch),
                                          step_on_batch=step_on_batch)


class ExponentialLR(TorchScheduler):
    """
    Example: ::

        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import ExponentialLR

        >>> # Example scheduler which multiplies the learning rate by 0.1 every epoch
        >>> scheduler = ExponentialLR(gamma=0.1)
        >>> trial = Trial(None, callbacks=[scheduler], metrics=['loss'], verbose=2).for_steps(10).run(1)

    Args:
        step_on_batch (bool): If True, step will be called on each training iteration rather than on each epoch

    See:
        `PyTorch ExponentialLR <http://pytorch.org/docs/master/optim.html#torch.optim.lr_scheduler.ExponentialLR>`_
    """
    def __init__(self, gamma, last_epoch=-1, step_on_batch=False):
        super(ExponentialLR, self).__init__(lambda opt:
                                            torch.optim.lr_scheduler.ExponentialLR(opt, gamma, last_epoch=last_epoch),
                                            step_on_batch=step_on_batch)


class CosineAnnealingLR(TorchScheduler):
    """
    Example: ::

        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import CosineAnnealingLR

        >>> # Example scheduler which uses cosine learning rate annealing - see PyTorch docs
        >>> scheduler = MultiStepLR(milestones=[30,80], gamma=0.1)
        >>> trial = Trial(None, callbacks=[scheduler], metrics=['loss'], verbose=2).for_steps(10).run(1)

    Args:
        step_on_batch (bool): If True, step will be called on each training iteration rather than on each epoch

    See:
        `PyTorch CosineAnnealingLR <http://pytorch.org/docs/master/optim.html#torch.optim.lr_scheduler.CosineAnnealingLR>`_
    """
    def __init__(self, T_max, eta_min=0, last_epoch=-1, step_on_batch=False):
        super(CosineAnnealingLR, self).__init__(lambda opt:
                                                torch.optim.lr_scheduler.CosineAnnealingLR(opt, T_max, eta_min=eta_min,
                                                                                           last_epoch=last_epoch),
                                                step_on_batch=step_on_batch)


class ReduceLROnPlateau(TorchScheduler):
    """
    Example: ::

        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import ReduceLROnPlateau

        >>> # Example scheduler which divides the learning rate by 10 on plateaus of 5 epochs without significant
        >>> # validation loss decrease, in order to stop overshooting the local minima. new_lr = lr * factor
        >>> scheduler = ReduceLROnPlateau(monitor='val_loss', factor=0.1, patience=5)
        >>> trial = Trial(None, callbacks=[scheduler], metrics=['loss'], verbose=2).for_steps(10).for_val_steps(10).run(1)

    Args:
        monitor (str): The name of the quantity in metrics to monitor. (Default value = 'val_loss')
        step_on_batch (bool): If True, step will be called on each training iteration rather than on each epoch

    See:
        `PyTorch ReduceLROnPlateau <http://pytorch.org/docs/master/optim.html#torch.optim.lr_scheduler.ReduceLROnPlateau>`_
    """
    def __init__(self,  monitor='val_loss', mode='min', factor=0.1, patience=10, verbose=False, threshold=1e-4,
                 threshold_mode='rel', cooldown=0, min_lr=0, eps=1e-8, step_on_batch=False):
        super(ReduceLROnPlateau, self).__init__(lambda opt:
                                                torch.optim.lr_scheduler.ReduceLROnPlateau(
                                                    opt, mode=mode, factor=factor, patience=patience, verbose=verbose,
                                                    threshold=threshold, threshold_mode=threshold_mode,
                                                    cooldown=cooldown, min_lr=min_lr, eps=eps), monitor=monitor,
                                                step_on_batch=step_on_batch)


class CyclicLR(TorchScheduler):
    """
    Example: ::

        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import CyclicLR

        >>> # Example scheduler which cycles the learning rate between 0.01 and 0.1
        >>> scheduler = CyclicLR(0.01, 0.1)
        >>> trial = Trial(None, callbacks=[scheduler], metrics=['loss'], verbose=2).for_steps(10).for_val_steps(10).run(1)

    Args:
        monitor (str): The name of the quantity in metrics to monitor. (Default value = 'val_loss')
        step_on_batch (bool): If True, step will be called on each training iteration rather than on each epoch

    See:
        `PyTorch ReduceLROnPlateau <http://pytorch.org/docs/master/optim.html#torch.optim.lr_scheduler.ReduceLROnPlateau>`_
    """
    def __init__(self,  base_lr, max_lr, monitor='val_loss', step_size_up=2000, step_size_down=None, mode='triangular',
                 gamma=1., scale_fn=None, scale_mode='cycle', cycle_momentum=True, base_momentum=0.8, max_momentum=0.9,
                 last_epoch=-1, step_on_batch=False):
        from distutils.version import LooseVersion
        version = torch.__version__ if str(torch.__version__) is torch.__version__ else "0.4.0"
        if LooseVersion(version) > LooseVersion("1.0.0"):  # CyclicLR is implemented
            super(CyclicLR, self).__init__(lambda opt:
                                                    torch.optim.lr_scheduler.CyclicLR(
                                                        opt, base_lr, max_lr, step_size_up=step_size_up,
                                                        step_size_down=step_size_down, mode=mode, gamma=gamma,
                                                        scale_fn=scale_fn, scale_mode=scale_mode,
                                                        cycle_momentum=cycle_momentum, base_momentum=base_momentum,
                                                        max_momentum=max_momentum, last_epoch=last_epoch),
                                           monitor=monitor, step_on_batch=step_on_batch)
        else:
            raise NotImplementedError('CyclicLR scheduler was not implemented in PyTorch versions less than 1.1.0. '
                                      'Update PyTorch or use the CyclicLR callback from an older Torchbearer version.')



import torchbearer
from torchbearer.callbacks import Callback
import torch
import warnings


class AggregatePredictions(Callback):
    def __init__(self):
        super(AggregatePredictions, self).__init__()
        self.predictions_list = []

    def on_step_validation(self, state):
        super(AggregatePredictions, self).on_step_validation(state)
        self.predictions_list.append(state[torchbearer.Y_PRED])

    def on_end_validation(self, state):
        try:
            if len(self.predictions_list) == 1 and type(self.predictions_list[0]) is torch.Tensor:
                state[torchbearer.FINAL_PREDICTIONS] = self.predictions_list[0]
            else:
                state[torchbearer.FINAL_PREDICTIONS] = torch.cat(self.predictions_list, 0)
        except:
            warnings.warn('Failed to format predictions as tensor, returning as list.')
            state[torchbearer.FINAL_PREDICTIONS] = self.predictions_list

    def on_end_epoch(self, state):
        super(AggregatePredictions, self).on_end_epoch(state)
        self.predictions_list = []



import torch
from unittest import TestCase

from mock import MagicMock, patch

import torchbearer
import torchbearer.callbacks.init as init


class TestWeightInit(TestCase):
    def test_modules_from_state(self):
        callback = init.WeightInit(targets=['Mock'])
        model = MagicMock()
        state = {torchbearer.MODEL: model}
        callback.on_init(state)
        self.assertTrue(model.modules.call_count == 1)

    def test_filter(self):
        mock = MagicMock()
        callback = init.WeightInit(initialiser=lambda m: m.test(), modules=[mock], targets=['Mock'])
        callback.on_init({})
        self.assertTrue(mock.test.call_count == 1)

        mock = MagicMock()
        callback = init.WeightInit(initialiser=lambda m: m.test(), modules=[mock], targets=['Not'])
        callback.on_init({})
        self.assertTrue(mock.test.call_count == 0)

    def test_module_list(self):
        mock = MagicMock()
        callback = init.WeightInit(initialiser=lambda m: m.test(), modules=[mock], targets=['Mock'])
        model = MagicMock()
        state = {torchbearer.MODEL: model}
        callback.on_init(state)
        self.assertTrue(model.modules.call_count == 0)


class TestSimpleInits(TestCase):
    @patch('torchbearer.callbacks.init.init')
    def test_kaiming(self, nn_init):
        callback = init.KaimingNormal(a=1, mode='test', nonlinearity='test2')
        mock = MagicMock()
        callback.initialiser(mock)
        nn_init.kaiming_normal_.assert_called_once_with(mock.weight.data, a=1, mode='test', nonlinearity='test2')

        callback = init.KaimingUniform(a=1, mode='test', nonlinearity='test2')
        mock = MagicMock()
        callback.initialiser(mock)
        nn_init.kaiming_uniform_.assert_called_once_with(mock.weight.data, a=1, mode='test', nonlinearity='test2')

    @patch('torchbearer.callbacks.init.init')
    def test_xavier(self, nn_init):
        callback = init.XavierNormal(gain=100)
        mock = MagicMock()
        callback.initialiser(mock)
        nn_init.xavier_normal_.assert_called_once_with(mock.weight.data, gain=100)

        callback = init.XavierUniform(gain=100)
        mock = MagicMock()
        callback.initialiser(mock)
        nn_init.xavier_uniform_.assert_called_once_with(mock.weight.data, gain=100)

    def test_bias(self):
        callback = init.ZeroBias()
        mock = MagicMock()
        callback.initialiser(mock)
        self.assertTrue(mock.bias.data.zero_.call_count == 1)


class TestLsuv(TestCase):
    def test_end_to_end(self):
        import numpy as np
        from torchbearer.callbacks.init import ZeroBias

        np.random.seed(7)
        torch.manual_seed(7)

        class Flatten(torch.nn.Module):
            def forward(self, x):
                return x.view(x.shape[0], -1)

        model = torch.nn.Sequential(
            torch.nn.Conv2d(1,1,1),
            Flatten(),
            torch.nn.Linear(4, 2),
        )

        state = {torchbearer.MODEL: model}
        data = torch.rand(2, 1, 2, 2)
        ZeroBias(model.modules()).on_init(state)  # LSUV expects biases to be zero
        init.LsuvInit(data).on_init(state)

        correct_conv_weight = torch.FloatTensor([[[[3.2236]]]])
        correct_linear_weight = torch.FloatTensor([[-0.3414, -0.5503, -0.4402, -0.4367],
                                                   [0.3425, -0.0697, -0.6646, 0.4900]])

        conv_weight = list(model.modules())[1].weight
        linear_weight = list(model.modules())[3].weight
        diff_conv = (conv_weight-correct_conv_weight) < 0.0001
        diff_linear = (linear_weight - correct_linear_weight) < 0.0001
        self.assertTrue(diff_conv.all().item())
        self.assertTrue(diff_linear.all().item())

    def test_break(self):
        import numpy as np
        from torchbearer.callbacks.init import ZeroBias

        np.random.seed(7)
        torch.manual_seed(7)

        model = torch.nn.Sequential(
            torch.nn.Conv2d(1,1,1),
        )

        with patch('torchbearer.callbacks.lsuv.LSUV.apply_weights_correction') as awc:
            state = {torchbearer.MODEL: model}
            data = torch.rand(2, 1, 2, 2)
            ZeroBias(model.modules()).on_init(state)  # LSUV expects biases to be zero
            init.LsuvInit(data, std_tol=1e-20, max_attempts=0, do_orthonorm=False).on_init(state)

            # torchbearer.callbacks.lsuv.apply_weights_correction = old_fun
            self.assertTrue(awc.call_count == 2)




import torchbearer
from torchbearer import Callback


class UnpackState(Callback):
    """Callback that unpacks a number of items from the state dictionary and passes it to the model forward as a separate
    dictionary (under the same keys).

    Note that torchbearer.X is always passed in the dictionary, if given as a key or not.

    Note that, if output_to_state is set then when the model outputs a dictionary, the main
    torchbearer state will be updated with this dictionary. So, if model outputs are
    {torchbearer.Y_PRED: 1, SOME_KEY: 10} then the main torchbearer state will look like:
    {..., torchbearer.Y_PRED: 1, SOME_KEY: 10, ...} after the forward pass. If Y_PRED is not a key in the model output
    dict or the output is not a dict then main state will look like: {..., torchbearer.Y_PRED: (model_output), ...}

    This is useful when using a DataParallel model where DataParallel cannot pass the main torchbearer state dictionary
    directly.

    Example: ::
        >>> import torch
        >>> from torch.nn import Module, DataParallel
        >>> import torchbearer
        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import UnpackState

        >>> class TestModel(Module):
        ...     def forward(self, x):
        ...         print(x)
        ...         return x

        >>> A_KEY = torchbearer.state_key('a_key')
        >>> unpacker = UnpackState(keys=[torchbearer.X, A_KEY])
        >>> t = Trial(TestModel(), callbacks=[unpacker])
        >>> t.state[A_KEY] = 'test'
        >>> t.with_train_data(torch.ones(10, 1), torch.ones(10, 1), batch_size=2, steps=1).run()
        {x: tensor([[1.],
                [1.]]), a_key: 'test'}
        [((1, None), {})]


    Args:
        keys (list / tuple): List of keys to unpack from state and pass to the model on forward
        output_to_state (bool): If True and torchbearer.Y_PRED in model_outputs, main torchbearer state will be updated
        (state.update(model_outputs)). If False then model outputs will be stored in the main state under
        torchbeaerer.Y_PRED.
    """
    def __init__(self, keys=None, output_to_state=True):
        super(UnpackState, self).__init__()
        self.output_to_state = output_to_state
        self.keys = keys if keys is not None else [torchbearer.X]
        if torchbearer.X not in self.keys:
            self.keys.insert(0, torchbearer.X)

    def on_sample(self, state):
        if self.keys != [torchbearer.X]:
            state[torchbearer.X] = {k: state[k] for k in self.keys}

    def on_sample_validation(self, state):
        self.on_sample(state)

    def on_forward(self, state):
        if self.output_to_state and isinstance(state[torchbearer.Y_PRED], dict):
            outputs = state[torchbearer.Y_PRED]
            state.update(outputs)
            state[torchbearer.Y_PRED] = outputs[torchbearer.Y_PRED] if torchbearer.Y_PRED in outputs else outputs

    def on_forward_validation(self, state):
        self.on_forward(state)



from unittest import TestCase

from mock import MagicMock, patch, ANY

import torch

import torchbearer
import torchbearer.callbacks.imaging as imaging


class TestClassAppearanceModel(TestCase):
    def test_targets_hot(self):
        callback = imaging.ClassAppearanceModel(nclasses=10, input_size=(1, 1, 1), target=5)
        targets_hot = callback._targets_hot({torchbearer.DEVICE: 'cpu'})
        self.assertTrue(targets_hot.sum() == 1)
        self.assertTrue(targets_hot[0][5] == 1)

    def test_targets_to_key(self):
        callback = imaging.ClassAppearanceModel(nclasses=10, input_size=(1, 1, 1), target=5)
        callback = callback.target_to_key('test')
        state = {torchbearer.DEVICE: 'cpu'}
        targets_hot = callback._targets_hot(state)
        self.assertTrue(targets_hot.sum() == 1)
        self.assertTrue(targets_hot[0][5] == 1)
        self.assertTrue(state['test'] == 5)

    @patch('torchbearer.callbacks.imaging.inside_cnns.torch.randint')
    def test_random_targets(self, randint):
        randint.return_value = torch.ones(1, 1) * 6
        callback = imaging.ClassAppearanceModel(nclasses=10, input_size=(1, 1, 1))
        state = {torchbearer.DEVICE: 'cpu'}
        targets_hot = callback._targets_hot(state)
        self.assertTrue(targets_hot.sum() == 1)
        self.assertTrue(targets_hot[0][6] == 1)

    def test_cam_loss(self):
        key = 'my_key'
        targets_hot = torch.FloatTensor([[0, 0, 0, 1, 0]]).ge(0.5)
        decay = 0.5
        loss = imaging.inside_cnns._cam_loss(key, targets_hot, decay)

        model = MagicMock
        model.input_image = torch.ones(5, 5) * 2

        state = {'my_key': torch.FloatTensor([[0, 0, 0, 100, 0]]), torchbearer.MODEL: model}
        res = loss(state)
        self.assertTrue(res.item() == -50)

    def test_cam_wrapper(self):
        model = MagicMock()
        wrapper = imaging.inside_cnns._CAMWrapper((3, 10, 10), model)

        self.assertTrue(wrapper.input_image.requires_grad)
        self.assertTrue(wrapper.input_image.shape == torch.Size([3, 10, 10]))
        wrapper('', 'state')
        self.assertTrue(model.call_count == 1)
        self.assertTrue(model.call_args[0][0].shape == torch.Size([1, 3, 10, 10]))
        self.assertTrue(model.call_args[0][1] == 'state')
        model.reset_mock()

        model.side_effect = lambda x: x
        wrapper('', 'state')
        self.assertTrue(model.call_count == 2)
        self.assertTrue(model.call)
        self.assertTrue(model.call_args[0][0].shape == torch.Size([1, 3, 10, 10]))

    @patch('torchbearer.callbacks.imaging.inside_cnns._cam_loss')
    @patch('torchbearer.callbacks.imaging.inside_cnns._CAMWrapper')
    @patch('torchbearer.callbacks.imaging.inside_cnns.torchbearer.Trial')
    def test_on_batch(self, _, wrapper, loss):
        wrapper().input_image = torch.nn.Parameter(torch.zeros(10))

        factory = MagicMock()
        callback = imaging.ClassAppearanceModel(nclasses=10, input_size=(1, 1, 1), optimizer_factory=factory)

        model = MagicMock()
        callback.on_batch({torchbearer.EPOCH: 0, torchbearer.MODEL: model, torchbearer.DEVICE: 'cpu', torchbearer.DATA_TYPE: torch.float32})

        loss.assert_called_once_with(torchbearer.PREDICTION, ANY, 0.01)

        self.assertTrue(model.eval.call_count == 1)
        self.assertTrue(model.train.call_count == 1)
        self.assertTrue(factory.call_count == 1)
        self.assertTrue(next(iter(factory.call_args[0][0])) is wrapper().input_image)

    def test_end_to_end(self):
        with torchbearer.no_grad():
            model = torch.nn.Linear(10, 5)
            callback = imaging.ClassAppearanceModel(nclasses=5, input_size=(10), steps=1)
            state = {torchbearer.EPOCH: 0, torchbearer.MODEL: model, torchbearer.DEVICE: 'cpu',
                     torchbearer.DATA_TYPE: torch.float32}
            callback.on_batch(state)



from __future__ import print_function
import sys
if sys.version_info[0] < 3:
    old_super = super

    def super(_, obj):
        return old_super(obj.__class__, obj)

import time
from torchbearer.callbacks import Callback
import torchbearer
from torchbearer.metrics import Metric

ON_START_TRAINING = 'on_start_training'
ON_START_EPOCH = 'on_start_epoch'
ON_SAMPLE = 'on_sample'
ON_FORWARD = 'on_forward'
ON_CRITERION = 'on_criterion'
ON_BACKWARD = 'on_backward'
ON_STEP_TRAINING = 'on_step_training'
ON_START_VALIDATION = 'on_start_validation'
ON_SAMPLE_VALIDATION = 'on_sample_validation'
ON_FORWARD_VALIDATION = 'on_forward_validaiton'
ON_CRITERION_VALIDATION = 'on_criterion_validation'
ON_STEP_VALIDATION = 'on_step_validation'
TRAIN_TIME = 'train_time'
TOTAL_TIME = 'total_time'
VALIDATION_TIME = 'validation_time'


class TimerMetric(Callback, Metric):
    def __init__(self, time_keys=()):
        """ Timer callback that aggregates timings for each stage of model execution
        """
        super(TimerMetric, self).__init__(name='timer')
        self.t0 = time.time()
        self.time_dict = {}
        # self.init_keys()
        self.batch_timer = _TimerMetric('t_batch')
        self.epoch_timer = _TimerMetric('t_epoch')
        self.train_timer = _TimerMetric('t_train')
        self.valid_timer = _TimerMetric('t_valid')
        self.total_timer = _TimerMetric('t_total')
        self.time_keys = time_keys
        self.added_callback = False

    def update_time(self, text, metric, state):
        self.time_dict[text] = metric.process(state)
        state[torchbearer.TIMINGS] = self.time_dict

    def process(self, *args):
        super(TimerMetric, self).process(*args)
        d_out = {key: self.time_dict[key] for key in self.time_keys if key in self.time_dict}
        return d_out

    def reset(self, state):
        super(TimerMetric, self).reset(state)
        if not self.added_callback:
            state[torchbearer.CALLBACK_LIST].append([self])
            self.added_callback = True

    def on_start(self, state):
        self.t0 = time.time()
        self.batch_timer.reset(state)
        self.epoch_timer.reset(state)
        self.train_timer.reset(state)
        self.valid_timer.reset(state)
        self.total_timer.reset(state)

    def on_start_training(self, state):
        super(TimerMetric, self).on_start_training(state)
        self.update_time(ON_START_TRAINING, self.batch_timer, state)
        self.update_time(ON_START_TRAINING, self.train_timer, state)

    def on_start_epoch(self, state):
        super(TimerMetric, self).on_start_epoch(state)
        self.update_time(ON_START_EPOCH, self.epoch_timer, state)

    def on_sample(self, state):
        super(TimerMetric, self).on_sample(state)
        self.update_time(ON_SAMPLE, self.batch_timer, state)

    def on_forward(self, state):
        super(TimerMetric, self).on_forward(state)
        self.update_time(ON_FORWARD, self.batch_timer, state)

    def on_criterion(self, state):
        super(TimerMetric, self).on_criterion(state)
        self.update_time(ON_CRITERION, self.batch_timer, state)

    def on_backward(self, state):
        super(TimerMetric, self).on_backward(state)
        self.update_time(ON_BACKWARD, self.batch_timer, state)

    def on_step_training(self, state):
        super(TimerMetric, self).on_step_training(state)
        self.update_time(ON_STEP_TRAINING, self.batch_timer, state)

    def on_start_validation(self, state):
        super(TimerMetric, self).on_start_validation(state)
        self.update_time(ON_START_VALIDATION, self.batch_timer, state)

    def on_sample_validation(self, state):
        super(TimerMetric, self).on_sample_validation(state)
        self.update_time(ON_SAMPLE_VALIDATION, self.batch_timer, state)

    def on_forward_validation(self, state):
        super(TimerMetric, self).on_forward_validation(state)
        self.update_time(ON_FORWARD_VALIDATION, self.batch_timer, state)

    def on_criterion_validation(self, state):
        super(TimerMetric, self).on_criterion_validation(state)
        self.update_time(ON_CRITERION_VALIDATION, self.batch_timer, state)

    def on_step_validation(self, state):
        super(TimerMetric, self).on_step_validation(state)
        self.update_time(ON_STEP_VALIDATION, self.batch_timer, state)

    def on_end_training(self, state):
        super(TimerMetric, self).on_end_training(state)
        self.valid_timer.reset(state)
        self.batch_timer.reset(state)
        self.update_time(TRAIN_TIME, self.train_timer, state)

    def on_end_epoch(self, state):
        super(TimerMetric, self).on_end_epoch(state)
        self.batch_timer.reset(state)
        self.train_timer.reset(state)

    def on_end(self, state):
        super(TimerMetric, self).on_end(state)
        self.update_time(TOTAL_TIME, self.total_timer, state)
        print(str(self.time_dict))

    def on_end_validation(self, state):
        super(TimerMetric, self).on_end_validation(state)
        self.update_time(VALIDATION_TIME, self.valid_timer, state)

    def get_timings(self):
        return self.time_dict


class _TimerMetric(Metric):
    def __init__(self, name):
        super(_TimerMetric, self).__init__(name)
        self.t = time.time()

    def process(self, *args):
        super(_TimerMetric, self).process(*args)
        dt = time.time() - self.t
        self.t = time.time()
        return dt

    def reset(self, state):
        super(_TimerMetric, self).reset(state)
        self.t = time.time()



import torch

import torchbearer
from torchbearer import cite
from torchbearer.callbacks import Callback


bibtex = """
@article{szegedy2015rethinking,
  title={Rethinking the inception architecture for computer vision. arXiv 2015},
  author={Szegedy, Christian and Vanhoucke, Vincent and Ioffe, Sergey and Shlens, Jonathon and Wojna, Zbigniew},
  journal={arXiv preprint arXiv:1512.00567},
  volume={1512},
  year={2015}
}
"""


@cite(bibtex)
class LabelSmoothingRegularisation(Callback):
    """Perform Label Smoothing Regularisation (LSR) on the targets during training. This involves converting the target
    to a one-hot vector and smoothing according to the value epsilon.

    .. note::

        Requires a multi-label loss, such as nn.BCELoss

    Example: ::

        >>> from torchbearer import Trial
        >>> from torchbearer.callbacks import LabelSmoothingRegularisation

        # Example Trial which does label smoothing regularisation
        >>> smoothing = LabelSmoothingRegularisation()
        >>> trial = Trial(None, criterion=nn.BCELoss(), callbacks=[smoothing], metrics=['acc'])

    Args:
        epsilon (float): The epsilon parameter from the paper
        classes (int): The number of target classes, not required if the target is already one-hot encoded
    """
    def __init__(self, epsilon, classes=-1):
        self.epsilon = epsilon
        self.classes = classes

    def to_one_hot(self, state):
        target = state[torchbearer.TARGET]

        if target.dim() == 1:
            target = target.unsqueeze(1)
            one_hot = torch.zeros_like(target).repeat(1, self.classes)
            one_hot.scatter_(1, target, 1)
            target = one_hot
        return target

    def on_sample(self, state):
        target = self.to_one_hot(state)
        target = (1 - self.epsilon) * target.float() + (self.epsilon / target.size(1))
        state[torchbearer.TARGET] = target

    def on_sample_validation(self, state):
        target = self.to_one_hot(state)
        state[torchbearer.TARGET] = target.float()



import sys
from unittest import TestCase

from mock import MagicMock, patch, ANY

import torchbearer
from torchbearer.callbacks import PyCM

import matplotlib.pyplot as plt  # Import so that it can be mocked
plt.ioff()


class TestHandlers(TestCase):
    @patch('matplotlib.pyplot')
    def test_to_pyplot(self, mock_pyplot):
        if sys.version_info[0] >= 3:
            import pycm

            handler = torchbearer.callbacks.pycm._to_pyplot(True, 'test {epoch}')

            y_true = [2, 0, 2, 2, 0, 1, 1, 2, 2, 0, 1, 2]
            y_pred = [0, 0, 2, 1, 0, 2, 1, 0, 2, 0, 2, 2]
            cm = pycm.ConfusionMatrix(y_true, y_pred)
            handler(cm, {torchbearer.EPOCH: 3})

            self.assertTrue(mock_pyplot.imshow.call_args[0][0].max() == 1)
            mock_pyplot.title.assert_called_once_with('test 3')

            handler = torchbearer.callbacks.pycm._to_pyplot(False)

            y_true = [2, 0, 2, 2, 0, 1, 1, 2, 2, 0, 1, 2]
            y_pred = [0, 0, 2, 1, 0, 2, 1, 0, 2, 0, 2, 2]
            cm = pycm.ConfusionMatrix(y_true, y_pred)
            handler(cm, {})

            self.assertTrue(mock_pyplot.imshow.call_args[0][0].max() > 1)


class TestPyCM(TestCase):
    def test_exception(self):
        if sys.version_info[0] < 3:
            self.assertRaises(Exception, PyCM)

    @patch('torchbearer.callbacks.pycm.EpochLambda')
    def test_make_cm(self, emock_lambda):
        if sys.version_info[0] >= 3:
            with patch('pycm.ConfusionMatrix') as confusion_mocktrix:
                confusion_mocktrix.return_value = 'test'
                handler = MagicMock()
                callback = PyCM(test=10).with_handler(handler)
                state = {torchbearer.METRIC_LIST: None}

                callback._add_metric(state)
                emock_lambda.assert_called_once_with('pycm', ANY, False)

                make_cm = emock_lambda.call_args[0][1]

                import torch

                y_pred = torch.rand(5, 2) / 2
                y_pred[:, 1] = 1
                y_true = MagicMock()

                make_cm(y_pred, y_true)

                self.assertTrue(y_true.cpu.call_count == 1)
                self.assertTrue(y_true.cpu().numpy.call_count == 1)
                confusion_mocktrix.assert_called_once_with(y_true.cpu().numpy(), ANY, test=10)
                self.assertTrue(confusion_mocktrix.call_args[0][1].sum() == 5)
                handler.assert_called_once_with('test', state)

    def test_on_train(self):
        if sys.version_info[0] >= 3:
            callback = PyCM().on_train()
            state = {torchbearer.METRIC_LIST: None}
            callback.on_start_training(state)
            self.assertTrue(state[torchbearer.METRIC_LIST] is not None)

    def test_on_val(self):
        if sys.version_info[0] >= 3:
            callback = PyCM().on_val()
            state = {torchbearer.METRIC_LIST: None, torchbearer.DATA: torchbearer.VALIDATION_DATA}
            callback.on_start_validation(state)
            self.assertTrue(state[torchbearer.METRIC_LIST] is not None)

    def test_on_test(self):
        if sys.version_info[0] >= 3:
            callback = PyCM().on_test()
            state = {torchbearer.METRIC_LIST: None, torchbearer.DATA: torchbearer.TEST_DATA}
            callback.on_start_validation(state)
            self.assertTrue(state[torchbearer.METRIC_LIST] is not None)

    def test_with_handler(self):
        if sys.version_info[0] >= 3:
            callback = PyCM()
            callback.with_handler('test')
            self.assertTrue('test' in callback._handlers)

    def test_to_state(self):
        if sys.version_info[0] >= 3:
            callback = PyCM()
            callback.to_state('test')
            out = {}
            callback._handlers[0]('cm', out)
            self.assertTrue('test' in out)
            self.assertTrue(out['test'] == 'cm')

    @patch('torchbearer.callbacks.pycm.print')
    def test_to_console(self, mock_print):
        if sys.version_info[0] >= 3:
            callback = PyCM()
            callback.to_console()
            callback._handlers[0]('cm', {})
            mock_print.assert_called_once_with('cm')

    def test_to_file(self):
        if sys.version_info[0] >= 3:
            callback = PyCM()
            callback.to_pycm_file('test {epoch}')
            cm = MagicMock()
            callback._handlers[0](cm, {torchbearer.EPOCH: 1})

            cm.save_stat.assert_called_once_with('test 1', address=True, overall_param=None, class_param=None,
                                                 class_name=None)

            callback = PyCM()
            callback.to_html_file('test {epoch}')
            cm = MagicMock()
            callback._handlers[0](cm, {torchbearer.EPOCH: 2})

            cm.save_html.assert_called_once_with('test 2', address=True, overall_param=None, class_param=None,
                                                 class_name=None, color=(0, 0, 0), normalize=False)

            callback = PyCM()
            callback.to_csv_file('test {epoch}')
            cm = MagicMock()
            callback._handlers[0](cm, {torchbearer.EPOCH: 3})

            cm.save_csv.assert_called_once_with('test 3', address=True, overall_param=None, class_param=None,
                                                class_name=None, matrix_save=True, normalize=False)

            callback = PyCM()
            callback.to_obj_file('test {epoch}')
            cm = MagicMock()
            callback._handlers[0](cm, {torchbearer.EPOCH: 4})

            cm.save_obj.assert_called_once_with('test 4', address=True, save_stat=False, save_vector=True)

    @patch('torchbearer.callbacks.pycm._to_pyplot')
    def test_to_pyplot(self, mock_to_pyplot):
        if sys.version_info[0] >= 3:
            PyCM().to_pyplot(True, 'test', 'test2')
            mock_to_pyplot.assert_called_once_with(normalize=True, title='test', cmap='test2')


import unittest

from mock import Mock, call

from torchbearer.metrics import RunningMean, Metric, RunningMetric, Mean, Std, Var

import torch


class TestVar(unittest.TestCase):
    def test_variance_dim(self):
        var = Var('test', dim=0)
        var.process(torch.Tensor([[1., 2.], [3., 4.]]))
        var.process(torch.Tensor([[4., 3.], [2., 1.]]))
        var.process(torch.Tensor([[1., 1.], [1., 1.]]))

        res = var.process_final()
        self.assertTrue(len(res) == 2)
        for m in res:
            self.assertTrue(abs(m - 1.6000) < 0.0001)


class TestStd(unittest.TestCase):
    def setUp(self):
        self._metric = Metric('test')
        self._metric.process = Mock()
        self._metric.process.side_effect = [torch.zeros(torch.Size([])),
                                            torch.FloatTensor([0.1, 0.2, 0.3]),
                                            torch.FloatTensor([0.4, 0.5, 0.6]),
                                            torch.FloatTensor([0.7, 0.8, 0.9]),
                                            torch.ones(torch.Size([]))]

        self._std = Std('test', unbiased=False)
        self._std.reset({})
        self._target = 0.31622776601684

    def test_train(self):
        self.setUp()
        self._std.train()
        for i in range(5):
            self._std.process(self._metric.process())
        result = self._std.process_final({})
        self.assertAlmostEqual(self._target, result, places=5)

    def test_validate(self):
        self.setUp()
        self._std.eval()
        for i in range(5):
            self._std.process(self._metric.process())
        result = self._std.process_final({})
        self.assertAlmostEqual(self._target, result, places=5)

    def test_precision_error(self):
        self.setUp()
        self._std.train()
        val = torch.tensor([0.55])
        for i in range(2):
            self._std.process(val)

        result = self._std.process_final({})
        self.assertEqual(0, result)

    def setUpMoreDims(self):
        self._metric = Metric('test')
        self._metric.process = Mock()
        self._metric.process.side_effect = [torch.zeros(torch.Size([])),
                                            torch.FloatTensor([[0.1, 0.2, 0.3], [1.1, 1.2, 1.3]]),
                                            torch.FloatTensor([[0.4, 0.5, 0.6], [1.4, 1.5, 1.6]]),
                                            torch.FloatTensor([[0.7, 0.8, 0.9], [1.7, 1.8, 1.9]]),
                                            torch.ones(torch.Size([]))]
        self._std = Std('test', unbiased=False)
        self._std.reset({})
        self._target = 0.57662804083742

    def test_more_dims(self):
        self.setUpMoreDims()
        for i in range(5):
            self._std.process(self._metric.process())
        result = self._std.process_final({})
        self.assertAlmostEqual(self._target, result, places=5)

    def test_std_dim(self):
        std = Std('test', dim=0)
        std.process(torch.Tensor([[1., 2.], [3., 4.]]))
        std.process(torch.Tensor([[4., 3.], [2., 1.]]))
        std.process(torch.Tensor([[1., 1.], [1., 1.]]))

        res = std.process_final()
        self.assertTrue(len(res) == 2)
        for m in res:
            self.assertTrue(abs(m - 1.2649) < 0.0001)


class TestMean(unittest.TestCase):
    def setUp(self):
        self._metric = Metric('test')
        self._metric.process = Mock()
        self._metric.process.side_effect = [torch.zeros(torch.Size([])),
                                            torch.FloatTensor([0.1, 0.2, 0.3]),
                                            torch.FloatTensor([0.4, 0.5, 0.6]),
                                            torch.FloatTensor([0.7, 0.8, 0.9]),
                                            torch.ones(torch.Size([]))]

        self._mean = Mean('test')
        self._mean.reset({})
        self._target = 0.5

    def test_train_dict(self):
        self.setUp()
        self._mean.train()
        for i in range(5):
            self._mean.process(self._metric.process())
        result = self._mean.process_final({})
        self.assertAlmostEqual(self._target, result, places=5)

    def test_validate_dict(self):
        self.setUp()
        self._mean.eval()
        for i in range(5):
            self._mean.process(self._metric.process())
        result = self._mean.process_final({})
        self.assertAlmostEqual(self._target, result, places=5)

    def setUpMoreDims(self):
        self._metric = Metric('test')
        self._metric.process = Mock()
        self._metric.process.side_effect = [torch.zeros(torch.Size([])),
                                            torch.FloatTensor([[0.1, 0.2, 0.3], [1.1, 1.2, 1.3]]),
                                            torch.FloatTensor([[0.4, 0.5, 0.6], [1.4, 1.5, 1.6]]),
                                            torch.FloatTensor([[0.7, 0.8, 0.9], [1.7, 1.8, 1.9]]),
                                            torch.ones(torch.Size([]))]
        self._mean = Mean('test')
        self._mean.reset({})
        self._target = 0.95

    def test_more_dims(self):
        self.setUpMoreDims()
        for i in range(5):
            self._mean.process(self._metric.process())
        result = self._mean.process_final({})
        self.assertAlmostEqual(self._target, result, places=5)

    def test_mean_dim(self):
        mean = Mean('test', dim=0)
        mean.process(torch.Tensor([[1., 2.], [3., 4.]]))
        mean.process(torch.Tensor([[4., 3.], [2., 1.]]))
        mean.process(torch.Tensor([[1., 1.], [1., 1.]]))

        res = mean.process_final()
        self.assertTrue(len(res) == 2)
        for m in res:
            self.assertTrue(abs(m - 2.0) < 0.0001)


class TestRunningMetric(unittest.TestCase):
    def setUp(self):
        self._metric = RunningMetric('test', batch_size=5, step_size=5)
        self._metric.reset({})
        self._metric._process_train = Mock(return_value=3)
        self._metric._step = Mock(return_value='output')

    def test_train_called_with_state(self):
        self._metric.train()
        self._metric.process({'test': -1})
        self._metric._process_train.assert_called_with({'test': -1})

    def test_cache_one_step(self):
        self._metric.train()
        for i in range(6):
            self._metric.process({})
        self._metric._step.assert_has_calls([call([3]), call([3, 3, 3, 3, 3])])

    def test_empty_methods(self):
        metric = RunningMetric('test')
        self.assertRaises(NotImplementedError, lambda: metric._step(['test']) is None)
        self.assertRaises(NotImplementedError, lambda: metric._process_train(['test']) is None)


class TestRunningMean(unittest.TestCase):
    def setUp(self):
        self._metric = Metric('test')
        self._mean = RunningMean('test')
        self._cache = [torch.Tensor([1.0]), torch.Tensor([1.5]), torch.Tensor([2.0])]
        self._target = 1.5

    def test_train(self):
        result = self._mean._process_train(torch.FloatTensor([1.0, 1.5, 2.0]))
        self.assertAlmostEqual(self._target, result, 3, 0.002)

    def test_step(self):
        result = self._mean._step(self._cache)
        self.assertEqual(self._target, result)

    def test_dims(self):
        mean = RunningMean('test', dim=0)
        cache = [mean._process_train(torch.Tensor([[1., 2.], [3., 4.]])),
                 mean._process_train(torch.Tensor([[4., 3.], [2., 1.]])),
                 mean._process_train(torch.Tensor([[1., 1.], [1., 1.]]))]

        res = mean._step(cache)
        self.assertTrue(len(res) == 2)
        for m in res:
            self.assertTrue(abs(m - 2.0) < 0.0001)



from torchbearer import Metric
from .metrics import *
from .wrappers import *
from .aggregators import *
from .decorators import *
from .roc_auc_score import *
from .primitives import *
from .timer import TimerMetric
from .default import DefaultAccuracy
from .lr import LR
